import pandas as pd
import numpy as np
import math
import pickle
from collections import Counter

from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedKFold
from sklearn import tree

import platform
from os import listdir
from os.path import isfile, join
from glob import glob
from pathlib import Path
import sys
import os
import copy
import traceback
from pathlib import Path

import traceback
import matplotlib.pyplot as plt
import graphviz

import birch
from predictor_advance_v1 import *
import utils

import sys
import traceback
import warnings
warnings.filterwarnings("ignore")

from ..config import Config

DEFAULT_CONFIG = Config()

def prepare_data(repo_name, directory, goal, config=DEFAULT_CONFIG):
    df_raw = pd.read_csv(directory + repo_name, sep=',')
    df_raw = df_raw.drop(columns=['dates'])
    last_col = utils.get_goal(goal, config)
    cols = list(df_raw.columns.values)
    cols.pop(cols.index(last_col))
    df_adjust = df_raw[cols+[last_col]]
    return df_adjust

# Cluster Driver
def cluster_driver(df,goal,print_tree = True, config=DEFAULT_CONFIG):
    X = df.apply(pd.to_numeric)
    cluster = birch.birch(branching_factor=config.branching_factor)
    cluster.fit(X)
    cluster_tree,max_depth = cluster.get_cluster_tree()
    if print_tree:
        cluster.show_clutser_tree()
    return cluster,cluster_tree,max_depth

def get_predicted(cluster_data, project_data, goal_num, metrics, month,seed, config=DEFAULT_CONFIG):
    try:
        train_data = pd.read_pickle(cluster_data + '/train_data.pkl')
        
        cluster,cluster_tree,max_depth = cluster_driver(train_data,goal_num,config)
        goal = get_goal(goal_num, config)
        print("GOAL NO: ",goal)
        conv_bell_df = pd.read_pickle('results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/default_bellwether.pkl')
        conv_bell_df = pd.DataFrame.from_dict(conv_bell_df, orient = 'index')
        # print("conv bell", conv_bell_df)
        conv_bell = conv_bell_df.median(axis = 0).idxmin()
        conv_bell_model_df = pd.read_pickle('results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/default_bellwether_models.pkl')
        conv_bell_clf = conv_bell_model_df[conv_bell]
        t_df = pd.DataFrame()
        for source_project in train_data.index:
            source_project_dataset = prepare_data(source_project, 
                                                    project_data, 
                                                    goal_num,
                                                    config)
            source_project_dataset = normalize(source_project_dataset)
            t_df = pd.concat([t_df,source_project_dataset])
        clf_global = DECART_test(t_df,metrics,month)
        
        # levels = [0,1,2] #TODO uncomment
        levels=[0,1]
        results = []
        for level in levels:
            bellwether_models = {}
            
            test_data = pd.read_pickle(cluster_data + '/test_data.pkl')
            
            bell_df = pd.read_csv(cluster_data + '/bellwether_level_' + str(level) + '.csv')
            bell_df.drop(['Unnamed: 0'], axis = 1, inplace = True)
            
            predicted_clusters = cluster.predict(test_data,level)
            
            test_data['predicted_clusters'] = predicted_clusters
            
            for test_project in test_data.index:
                result = []
                result.append(test_project)
                result.append(level)
                
                predicted_cluster = test_data.loc[test_project,'predicted_clusters']
                # print("predicted clusters: ",predicted_cluster)
                # print("bell df", bell_df)
                predicted_cluster_bellwether = bell_df[bell_df['id'] == predicted_cluster].bellwether.values[0]
                
                if predicted_cluster_bellwether not in bellwether_models.keys():
                    bellwether_models[predicted_cluster_bellwether] = None
                
                testing_project_dataset = prepare_data(test_project, 
                                                    project_data, 
                                                    goal_num,
                                                    config)
                testing_project_dataset = normalize(testing_project_dataset)


                trainData, testData = df_split_month(testing_project_dataset, month)
                
                y_train = trainData[goal]
                X_train = trainData.drop([goal], axis = 1)
                
                y_test = testData[goal]
                X_test = testData.drop([goal], axis = 1)
                
                train_df = X_train
                train_df[goal] = y_train
                
                
                
                clf_test = DECART_test(train_df,metrics,month)
                # print("XTEST:  ")
                # print( X_test)
                # if not X_test:
                #     continue
                test_predict = clf_test.predict(X_test)
                test_actual = y_test.values
                
                mre = mre_calc(test_predict, test_actual)
                sa = sa_calc(test_predict, test_actual, y_train)
                
                result.append(mre)
                result.append(sa)

                bellwether_dataset = prepare_data(predicted_cluster_bellwether, 
                                                project_data, 
                                                goal_num,
                                                config)
                if level == 2:
                    with open(path + '/' + str(predicted_cluster) + '/goal_' +  str(goal_num) + '_models.pkl', 'rb') as handle:
                        models = pickle.load(handle)
                    with open(path + '/' + str(predicted_cluster) + '/goal_' +  str(goal_num) + '_features.pkl', 'rb') as handle:
                        features = pickle.load(handle)
                else:
                    with open(path + '/level_' + str(level) + '/cluster_' +  str(predicted_cluster) + '_models.pkl', 'rb') as handle:
                        models = pickle.load(handle)
                    with open(path + '/level_' + str(level) + '/cluster_' +  str(predicted_cluster) + '_features.pkl', 'rb') as handle:
                        features = pickle.load(handle)
                
                print(level, predicted_cluster_bellwether, predicted_cluster)
                clf_bellwether = models[predicted_cluster_bellwether]
                clf_bellwether_cols = features[predicted_cluster_bellwether][:-1]
                
                test_predict = clf_bellwether.predict(X_test[clf_bellwether_cols])
                test_actual = y_test.values
                
                mre = mre_calc(test_predict, test_actual)
                sa = sa_calc(test_predict, test_actual, y_train)
                
                result.append(mre)
                result.append(sa)

                test_predict = conv_bell_clf.predict(X_test)
                test_actual = y_test.values
                
                mre = mre_calc(test_predict, test_actual)
                sa = sa_calc(test_predict, test_actual, y_train)

                result.append(mre)
                result.append(sa)
        
                test_predict = clf_global.predict(X_test)
                test_actual = y_test.values
                
                mre = mre_calc(test_predict, test_actual)
                sa = sa_calc(test_predict, test_actual, y_train)

                result.append(mre)
                result.append(sa)
            
                results.append(result)
        
        results_df = pd.DataFrame(results, columns = ['project','level','self_mre','self_sa','bellwether_mre','bellwether_sa','conv_bell_mre','conv_bell_sa','global_mre','global_sa'])
        # results_df = pd.DataFrame(results, columns = ['project','level','bellwether_mre','bellwether_sa']) #todo add only bellwethers
                
        return results_df 
    except Exception as e:
        # Handle the exception here
        print("An error occurred in get_predicted:", e)
        traceback.print_exc()


def performance_calculator(config=DEFAULT_CONFIG):
    month = config.month
    no_goals = config.no_goals
    seed = config.seed

    for i in range(no_goals):
        goal = get_goal(i, config)
        path  = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal
        results_df = get_predicted(cluster_data = path, 
                    project_data = "data/data_use/", 
                    goal_num = i, 
                    metrics = 0, 
                    month = month,
                    seed= seed,
                    config=config)
        results_df.to_csv(path + '/results_tuned_final_sa.csv')


if __name__ == "__main__":
    month = 12
    no_goals= 7
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        for i in range(no_goals):
            goal = get_goal(i, DEFAULT_CONFIG)
            print("goal:", goal)
            path  = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal
            results_df = get_predicted(cluster_data = path, 
                        project_data = 'data/data_use/', 
                        goal_num = i, 
                        metrics = 0, 
                        month = month,
                        seed= seed,
                        config=DEFAULT_CONFIG)
            results_df.to_csv(path + '/results_tuned_final_sa.csv')


import os
import pandas as pd

import utils

from ..config import Config

DEFAULT_CONFIG = Config()

def collect_bellwether(config=DEFAULT_CONFIG):
    month = config.month
    no_goals= config.no_goals
    seed = config.seed

    output_folder = "bellwethers/"

    for i in range(no_goals):
        goal = utils.get_goal(i, config)

        res_df_0 = pd.DataFrame()
        res_df_1 = pd.DataFrame()

        result_source = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/'
        bellwether_0 = pd.read_csv(os.path.join(result_source, "bellwether_level_0.csv"))
        bellwether_1 = pd.read_csv(os.path.join(result_source, "bellwether_level_1.csv"))

        bellwether_0["seed"] = seed
        bellwether_1["seed"] = seed

        res_df_0 = res_df_0.append(bellwether_0, ignore_index=True)
        res_df_1 = res_df_1.append(bellwether_1, ignore_index=True)
        
        if not os.path.exists("meta_data/old_new_names_mapping.csv"):
            print("Old names - New names mapping missing. Exiting ...")
            exit(0)

        mapping_df = pd.read_csv("meta_data/old_new_names_mapping.csv")

        res_df_0 = pd.merge(res_df_0, mapping_df, how='left', left_on='bellwether', right_on='new')
        res_df_1 = pd.merge(res_df_1, mapping_df, how='left', left_on='bellwether', right_on='new')

        res_df_0.drop(columns=['Unnamed: 0', "id", "new"], inplace=True)
        res_df_0.rename(columns={"old":"bellwether_project"}, inplace=True)

        res_df_1.drop(columns=['Unnamed: 0', "id", "new"], inplace=True)
        res_df_1.rename(columns={"old":"bellwether_project"}, inplace=True)

        res_df_0.sort_values(by="bellwether", inplace=True)
        res_df_1.sort_values(by="bellwether", inplace=True)

        res_df_0.to_csv(os.path.join(output_folder,goal+"_level_0.csv"), index=False)
        res_df_1.to_csv(os.path.join(output_folder,goal+"_level_1.csv"), index=False)


if __name__ == "__main__":
    seeds={
        'seeds1-1':[1836],
        'seeds2-1':[8023],
        'seeds3-1':[6241],
        'seeds4-1':[3578],
        'seeds5-1':[5347],
        'seeds6-1':[2916],
        'seeds7-1':[1021],
        'seeds8-1':[9562],
        'seeds9-1':[8115],
        'seeds10-1':[6931],
        'seeds1-2':[4629],
        'seeds2-2':[9774],
        'seeds3-2':[1592],
        'seeds4-2':[8865],
        'seeds5-2':[7264],
        'seeds6-2':[6489],
        'seeds7-2':[3745],
        'seeds8-2':[5207],
        'seeds9-2':[2653],
        'seeds10-2':[4815],
    }

    month = 12
    no_goals= 7

    output_folder = "bellwethers/"

    for i in range(no_goals):
        goal = utils.get_goal(i, config=DEFAULT_CONFIG)

        res_df_0 = pd.DataFrame()
        res_df_1 = pd.DataFrame()

        for _, seed in seeds.items():
            seed = seed[0]
            result_source = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/'
            bellwether_0 = pd.read_csv(os.path.join(result_source, "bellwether_level_0.csv"))
            bellwether_1 = pd.read_csv(os.path.join(result_source, "bellwether_level_1.csv"))

            bellwether_0["seed"] = seed
            bellwether_1["seed"] = seed

            res_df_0 = res_df_0.append(bellwether_0, ignore_index=True)
            res_df_1 = res_df_1.append(bellwether_1, ignore_index=True)
        
        mapping_df = pd.read_csv("old_new_names_mapping.csv")

        res_df_0 = pd.merge(res_df_0, mapping_df, how='left', left_on='bellwether', right_on='new')
        res_df_1 = pd.merge(res_df_1, mapping_df, how='left', left_on='bellwether', right_on='new')

        res_df_0.drop(columns=['Unnamed: 0', "id", "new"], inplace=True)
        res_df_0.rename(columns={"old":"bellwether_project"}, inplace=True)

        res_df_1.drop(columns=['Unnamed: 0', "id", "new"], inplace=True)
        res_df_1.rename(columns={"old":"bellwether_project"}, inplace=True)

        res_df_0.sort_values(by="bellwether", inplace=True)
        res_df_1.sort_values(by="bellwether", inplace=True)

        res_df_0.to_csv(os.path.join(output_folder,goal+"_level_0.csv"), index=False)
        res_df_1.to_csv(os.path.join(output_folder,goal+"_level_1.csv"), index=False)









import pandas as pd
import numpy as np

import utils

import platform
from os import listdir
from os.path import isfile, join
from glob import glob
from pathlib import Path
import sys
import os
import copy
import traceback
import timeit
import random
from ..config import Config

DEFAULT_CONFIG = Config()

def calculate_median(repo_pool, path, goal, config):
    feature_selected = []
    repos = []
    for repo in sorted(repo_pool):
        try:
            data = utils.data_goal_arrange(repo, path, goal, config)
            data = data.iloc[:,:-1]
            # print("data values: ",data['license'][0])
            # print("data: ", data.median())
            data_median = data.median().values.tolist()
            # data_median.append(data['license'][0])
            # print("data median: ",data_median)
            feature_selected.append(data_median)
            repos.append(repo)
        except Exception as e:
            print(e)
            continue
    return feature_selected, repos

def attribute_selector(config=DEFAULT_CONFIG):
    repo_pool = []
    path = "data/data_use/"
    no_goals= config.no_goals
    seed = config.seed

    for filename in os.listdir(path):
        if not filename.startswith('.'):
            repo_pool.append(os.path.join(filename))

    for _goal in range(no_goals):
        goal = utils.get_goal(_goal, config)
        print("goal: ",goal)
        df_cols = utils.data_goal_arrange(repo_pool[0], path, _goal, config).columns[:-1]
        feature_selected, repos = calculate_median(repo_pool, path, _goal)
        median_df = pd.DataFrame(feature_selected, columns = df_cols, index = repos)
        median_df.to_csv('results/attribute/'+str(seed)+'/data_attribute_' + goal + '.csv')


if __name__ == '__main__':
    repo_pool = []
   
    path = 'data/data_use/'
    no_goals= 7
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        for filename in os.listdir(path):
            if not filename.startswith('.'):
                repo_pool.append(os.path.join(filename))

        for _goal in range(no_goals):
            goal = utils.get_goal(_goal, DEFAULT_CONFIG)
            print("goal: ",goal)
            df_cols = utils.data_goal_arrange(repo_pool[0], path, _goal, DEFAULT_CONFIG).columns[:-1]
            print("df_cols: ",len(df_cols))
            feature_selected, repos = calculate_median(repo_pool, path, _goal)
            median_df = pd.DataFrame(feature_selected, columns = df_cols, index = repos)
            median_df.to_csv('results/attribute/'+str(seed)+'/data_attribute_' + goal + '.csv')



import pandas as pd
import numpy as np
import math
import pickle
from collections import Counter

from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedKFold

import platform
from os import listdir
from os.path import isfile, join
from glob import glob
from pathlib import Path
import sys
import os
import copy
import traceback
from pathlib import Path


import birch_bellwether_p_CFS as birch_bell
import birch
import utils

import sys
import traceback
import warnings
warnings.filterwarnings("ignore")

from ..config import Config

DEFAULT_CONFIG = Config()

# Birch Cluster Creator
def cluster_driver(df,print_tree = True, config=DEFAULT_CONFIG):
    X = df.apply(pd.to_numeric)
    cluster = birch.birch(branching_factor=config.branching_factor)
        #X.set_index('Project Name',inplace=True)
    cluster.fit(X)
    cluster_tree,max_depth = cluster.get_cluster_tree()
        #cluster_tree = cluster.model_adder(cluster_tree)
    if print_tree:
        cluster.show_clutser_tree()
    return cluster,cluster_tree,max_depth

def build_BIRCH(attr_df, config=DEFAULT_CONFIG):
    cluster,cluster_tree,_ = cluster_driver(attr_df, config)
    return cluster,cluster_tree

def get_clusters(data_source):
    if platform.system() == 'Darwin' or platform.system() == 'Linux':
        _dir = data_source + '/'
    else:
        _dir = data_source + '\\'

    clusters = [(join(_dir, f)) for f in listdir(_dir) if Path(join(_dir, f)).is_dir()]
    return clusters

def get_bellwether(df):
    bellwether = df.median_score.idxmin()
    return bellwether


def calculate_performance(data_source,clusters,path, goal,month, config=DEFAULT_CONFIG):
    df_train = pd.read_pickle(data_source + '/train_data.pkl')
    cluster,cluster_tree = build_BIRCH(df_train, config)
    cluster_ids = []
    cluster_structure = {}
    size = {}
    for key in cluster_tree:
        if cluster_tree[key].depth != None:
            cluster_ids.append(key)
            if cluster_tree[key].depth not in cluster_structure.keys():
                cluster_structure[cluster_tree[key].depth] = {}
            cluster_structure[cluster_tree[key].depth][key] = cluster_tree[key].parent_id
            size[key] = cluster_tree[key].size
    count = 0
    score = []
    score_med = []
    cluster_info = {}
    for cluster in clusters:
        if cluster in ['level_1','level_0']:
            continue
        perf_df = pd.read_pickle(data_source + str(cluster) + '/goal_' + str(goal) + '.pkl')
        avg_score = {}
        for s_project in perf_df.keys():
            if s_project not in avg_score.keys():
                avg_score[s_project] = []
            for d_project in perf_df[s_project].keys():
                avg_score[s_project].append(perf_df[s_project][d_project])
        perf_df = pd.DataFrame.from_dict(avg_score, orient = 'index')
        perf_df['median_score'] = perf_df.median(axis = 1)
        best_project = get_bellwether(perf_df) # get the best project that minimum median score
        score_med.append([cluster,best_project])
    score_df = pd.DataFrame(score_med, columns = ['id','bellwether'])
    score_df = score_df.sort_values('id')
    score_df.to_csv(data_source + '/bellwether_level_2.csv')
    level_1_bellwethers = {}

    # for cluster in cluster_structure[2].keys():
    #     if cluster_structure[2][cluster] not in level_1_bellwethers.keys():
    #         level_1_bellwethers[cluster_structure[2][cluster]] = []
    #     level_1_bellwethers[cluster_structure[2][cluster]].append(score_df[score_df['id'] == str(cluster)].bellwether.values.tolist()[0])
    
    for cluster in cluster_structure[1].keys():
        if cluster not in level_1_bellwethers.keys():
            level_1_bellwethers[cluster] = list(df_train.loc[cluster_tree[cluster].data_points].index)
    score_med = []
    
    data_path = Path(data_source + 'level_1')
    if not data_path.is_dir():
        os.makedirs(data_path)
    
    for key in  level_1_bellwethers.keys():
        sub_cluster_bellwethers = level_1_bellwethers[key]
        bell = birch_bell.Bellwether(path,df_train, goal,month, config=config)
        results = bell.bellwether(sub_cluster_bellwethers,sub_cluster_bellwethers)
        final_score = results[0]
        final_models = results[1]
        final_features = results[2]
        with open(data_source + 'level_1/cluster_'  + str(key) + '_performance.pkl', 'wb') as handle:
            pickle.dump(final_score, handle, protocol=pickle.HIGHEST_PROTOCOL)  
            
        with open(data_source + 'level_1/cluster_'  + str(key) + '_models.pkl', 'wb') as handle:
            pickle.dump(final_models, handle, protocol=pickle.HIGHEST_PROTOCOL)

        with open(data_source + 'level_1/cluster_'  + str(key) + '_features.pkl', 'wb') as handle:
            pickle.dump(final_features, handle, protocol=pickle.HIGHEST_PROTOCOL)
    
    level_1_bellwethers = {}
    score_med = []
    for cluster in cluster_structure[1].keys():
        cluster_perf = pd.read_pickle(data_source + 'level_1/cluster_'  + str(cluster) + '_performance.pkl')
        level_1_avg_score = {}
        for s_project in cluster_perf.keys():
            if s_project not in level_1_avg_score.keys():
                level_1_avg_score[s_project] = []
            for d_project in cluster_perf[s_project].keys():
                level_1_avg_score[s_project].append(cluster_perf[s_project][d_project])
        cluster_perf_df = pd.DataFrame.from_dict(level_1_avg_score, orient = 'index')
        cluster_perf_df['median_score'] = cluster_perf_df.median(axis = 1)
        best_project = get_bellwether(cluster_perf_df)
        score_med.append([cluster,best_project])
    score_df = pd.DataFrame(score_med, columns = ['id','bellwether'])
    score_df = score_df.sort_values('id')
    score_df.to_csv(data_source + '/bellwether_level_1.csv')
    sub_cluster_bellwethers = score_df.bellwether.values.tolist()
    results = bell.bellwether(sub_cluster_bellwethers,sub_cluster_bellwethers)
    final_score = results[0]
    final_models = results[1]
    final_features = results[2]
    
    data_path = Path(data_source + 'level_0')
    if not data_path.is_dir():
        os.makedirs(data_path)
    
    with open(data_source + '/level_0/cluster_0_performance.pkl', 'wb') as handle:
        pickle.dump(final_score, handle, protocol=pickle.HIGHEST_PROTOCOL)
    
    with open(data_source + '/level_0/cluster_0_models.pkl', 'wb') as handle:
        pickle.dump(final_models, handle, protocol=pickle.HIGHEST_PROTOCOL)

    with open(data_source + '/level_0/cluster_0_features.pkl', 'wb') as handle:
        pickle.dump(final_features, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
    level_0_avg_score = {}
    score_med = []
    for s_project in final_score.keys():
        if s_project not in level_0_avg_score.keys():
            level_0_avg_score[s_project] = []
        for d_project in final_score[s_project].keys():
            level_0_avg_score[s_project].append(final_score[s_project][d_project])  
    cluster_perf_df = pd.DataFrame.from_dict(level_0_avg_score, orient = 'index')
    cluster_perf_df['median_score'] = cluster_perf_df.median(axis = 1)
    best_project = get_bellwether(cluster_perf_df)
    score_med.append([0,best_project])
    score_df = pd.DataFrame(score_med, columns = ['id','bellwether'])
    score_df = score_df.sort_values('id')
    score_df.to_csv(data_source + '/bellwether_level_0.csv')


def find_bellwether(config=DEFAULT_CONFIG):
    month = config.month
    no_goals= config.no_goals
    seed = config.seed

    for i in range(no_goals):
        goal = utils.get_goal(i, config)
        path = "data/data_use/"
        data_source = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/'
        if platform.system() == 'Darwin' or platform.system() == 'Linux':
            _dir = data_source + '/'
        else:
            _dir = data_source + '\\'
        clusters = [f for f in listdir(_dir) if Path(join(_dir, f)).is_dir()]
        calculate_performance(data_source,clusters,path,i,month, config)  


if __name__ == "__main__":
    month = 12
    no_goals= 7
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        for i in range(no_goals):
            goal = utils.get_goal(i, DEFAULT_CONFIG)
            path = 'data/data_use/'
            data_source = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/'
            if platform.system() == 'Darwin' or platform.system() == 'Linux':
                _dir = data_source + '/'
            else:
                _dir = data_source + '\\'
            clusters = [f for f in listdir(_dir) if Path(join(_dir, f)).is_dir()]
            print("Clusters: ")
            print(clusters)
            calculate_performance(data_source,clusters,path,i,month, DEFAULT_CONFIG)  




import numpy as np
import random
from sklearn.tree import DecisionTreeRegressor
from sklearn import tree
# import matplotlib.pyplot as plt
from utils import *

def de(fun_opt, metrics, bounds, mut=0.5, crossp=0.5, popsize=20, itrs=10):
    dimensions = len(bounds)
    pop = np.random.rand(popsize, dimensions)
    min_b, max_b = np.asarray(bounds)[:,0], np.asarray(bounds)[:,1]
    diff = np.fabs(min_b - max_b)
    print(diff)
    # pop_denorm = min_b + pop * diff
    pop_denorm = np.rint(min_b + pop * diff).astype(np.int)
    fitness = np.asarray([fun_opt(*ind) for ind in pop_denorm])
    best_idx = np.argmin(fitness)
    best = pop_denorm[best_idx]
    for i in range(itrs):
        for j in range(popsize):
            idxs = [idx for idx in range(popsize) if idx != j]
            a, b, c = pop[np.random.choice(idxs, 3, replace=False)]
            mutant = np.clip(a + mut * (b - c), 0, 1)
            cross_points = np.random.rand(dimensions) < crossp
            if not np.any(cross_points):
                cross_points[np.random.randint(0, dimensions)] = True
            trial = np.where(cross_points, mutant, pop[j])
            # trial_denorm = min_b + trial * diff
            trial_denorm = np.rint(min_b + trial * diff).astype(np.int)
            f = fun_opt(*trial_denorm)
            if metrics == 0:
                if f < fitness[j]:  ####### MRE
                    fitness[j] = f
                    pop[j] = trial
                    if f < fitness[best_idx]:  ####### MRE
                        best_idx = j
                        best = trial_denorm
            if metrics == 1:
                if f > fitness[j]:  ####### SA
                    fitness[j] = f
                    pop[j] = trial
                    if f > fitness[best_idx]:  ####### SA
                        best_idx = j
                        best = trial_denorm
        return fitness[best_idx], best


def flash(train_input, train_actual_effort, test_input, test_actual_effort, metrics, pop_size):
    def convert(index):  # 12 12 20
        a = int(index / 240 + 1)
        b = int(index % 240 / 20 + 1)
        c = int(index % 20 + 2)
        return a, b, c

    def convert_lr(index):  # 30 2 2 100
        a = int(index / 400 + 1)
        b = int(index % 400 / 200 + 1)
        c = int(index % 200 / 100 + 1)
        d = int(index % 100 + 1)
        return a, b, c, d

    all_case = set(range(0, 2880))
    modeling_pool = random.sample(all_case, pop_size)

    List_X = []
    List_Y = []

    for i in range(len(modeling_pool)):
        temp = convert(modeling_pool[i])
        List_X.append(temp)
        model = DecisionTreeRegressor(max_depth=temp[0], min_samples_leaf=temp[1], min_samples_split=temp[2])
        model.fit(train_input, train_actual_effort)
        test_predict_effort = model.predict(test_input)
        test_predict_Y = test_predict_effort
        test_actual_Y = test_actual_effort.values

        if metrics == 0:
            List_Y.append(mre_calc(test_predict_Y, test_actual_Y))  ######### for MRE
        if metrics == 1:
            List_Y.append(sa_calc(test_predict_Y, test_actual_Y, train_actual_effort))  ######### for SA

    remain_pool = all_case - set(modeling_pool)
    test_list = []
    for i in list(remain_pool):
        test_list.append(convert(i))

    upper_model = DecisionTreeRegressor()
    life = 20

    while len(List_X) < 201 and life > 0:  # eval_number
        upper_model.fit(List_X, List_Y)
        candidate = random.sample(test_list, 1)
        test_list.remove(candidate[0])
        candi_pred_value = upper_model.predict(candidate)
        if metrics == 0:
            if candi_pred_value < np.median(List_Y):  ######### for MRE
                List_X.append(candidate[0])
                candi_config = candidate[0]
                candi_model = DecisionTreeRegressor(max_depth=candi_config[0], min_samples_leaf=candi_config[1],
                                                    min_samples_split=candi_config[2])
                candi_model.fit(train_input, train_actual_effort)
                candi_pred_Y = candi_model.predict(test_input)
                candi_actual_Y = test_actual_effort.values

                List_Y.append(mre_calc(candi_pred_Y, candi_actual_Y))  ######### for MRE

            else:
                life -= 1

        if metrics == 1:
            if candi_pred_value > np.median(List_Y):  ######### for SA
                List_X.append(candidate[0])
                candi_config = candidate[0]
                candi_model = DecisionTreeRegressor(max_depth=candi_config[0], min_samples_leaf=candi_config[1],
                                                    min_samples_split=candi_config[2])
                candi_model.fit(train_input, train_actual_effort)
                candi_pred_Y = candi_model.predict(test_input)
                candi_actual_Y = test_actual_effort.values

                List_Y.append(sa_calc(candi_pred_Y, candi_actual_Y, train_actual_effort))  ######### for SA

            else:
                life -= 1

    # temp_tree = candi_model
    # tree.plot_tree(temp_tree, feature_names=list(train_input.columns.values))
    # plt.show()

    if metrics == 0:
        return np.min(List_Y)  ########## min for MRE
    if metrics == 1:
        return np.max(List_Y)  ########## min for SA



from __future__ import print_function, division
from collections import Counter
import numpy as np
from sklearn.svm import SVC
from sklearn.metrics import classification_report
from sklearn.neighbors import KNeighborsClassifier
from sklearn.cluster import KMeans
from sklearn.naive_bayes import GaussianNB
from sklearn.linear_model import LogisticRegression
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
import traceback
from utils import *
import pdb
__author__ = 'Suvodeep'


class Learners(object):
  def __init__(self, clf, train_X, train_Y, predict_X, predict_Y, goal):
    """

    :param clf: classifier, SVM, etc...
    :param train_X: training data, independent variables.
    :param train_Y: training labels, dependent variables.
    :param predict_X: testing data, independent variables.
    :param predict_Y: testingd labels, dependent variables.
    :param goal: the objective of your tuning, F, recision,....
    """
    self.train_X = train_X.values.tolist()
    self.train_Y = train_Y.values.tolist()
    self.predict_X = predict_X.values
    self.predict_Y = predict_Y.values
    self.goal = goal
    self.param_distribution = self.get_param()
    self.learner = clf
    self.confusion = None
    self.params = None
    self.scores = None

  def learn(self, F, **kwargs):
    """
    :param F: a dict, holds all scores, can be used during debugging
    :param kwargs: a dict, all the parameters need to set after tuning.
    :return: F, scores.
    """
    try:
      self.learner.set_params(**kwargs)
      clf = self.learner.fit(self.train_X, self.train_Y)
      predict_result = []
      predict_Y = []
      # for predict_X, actual in zip(self.predict_X, self.predict_Y):
      #   try:
      #     _predict_result = clf.predict(predict_X.reshape(1, -1))
      #     predict_result.append(_predict_result[0])
      #     predict_Y.append(actual)
      #   except:
      #     print("one pass")
      #     continue
      predict_result = clf.predict(self.predict_X)
      if self.goal == 0:
          self.scores =  {'0':mre_calc(predict_result, self.predict_Y)}
      if self.goal == 1:
          self.scores =  {'1':sa_calc(predict_result, self.predict_Y, self.train_Y)}
      self.params = kwargs
      return self.scores
    except Exception as e:
        # Handle the exception here
        print("An error occurred in learn:", e)
        print(self.train_X, self.train_Y)
        traceback.print_exc()
        return 


  def _Abcd(self, predicted, actual, F):
    """

    :param predicted: predicted results(labels)
    :param actual: actual results(labels)
    :param F: previously got scores
    :return: updated scores.
    """
    def calculate(scores):
      for i, v in enumerate(scores):
        F[uni_actual[i]] = F.get(uni_actual[i], []) + [v]
      freq_actual = [count_actual[one] / len(actual) for one in uni_actual]
      F["mean"] = F.get("mean", []) + [np.mean(scores)]
      F["mean_weighted"] = (F.get("mean_weighted", []) + [
        np.sum(np.array(scores) * np.array(freq_actual))])
      return F

    def micro_cal(goal="Micro_F"):
      TP, FN, FP = 0, 0, 0
      for each in confusion_matrix_all_class:
        TP += each.TP
        FN += each.FN
        FP += each.FP
      PREC = TP / (TP + FP)
      PD = TP / (TP + FN)
      F[goal] = F.get(goal, []) + [round(2.0 * PREC * PD / (PD + PREC),3)]
      return F

    def macro_cal(goal="Macro_F"):
      PREC_sum, PD_sum = 0, 0
      for each in confusion_matrix_all_class:
        PD_sum += each.stats()[0]
        PREC_sum += each.stats()[2]
      PD_avg = PD_sum / len(uni_actual)
      PREC_avg = PREC_sum / len(uni_actual)
      F[goal] = F.get(goal, []) + [
        round(2.0 * PREC_avg * PD_avg / (PREC_avg + PD_avg),3)]
      return F

    _goal = {0: "PD", 1: "PF", 2: "PREC", 3: "ACC",
             4: "F", 5: "G", 6: "Macro_F", 7: "Micro_F"}
    abcd = ABCD(actual, predicted)
    # print(actual)
    uni_actual = list(set(actual))
    #print(uni_actual)
    count_actual = Counter(actual)
    if "Micro" in self.goal or "Macro" in self.goal:
      confusion_matrix_all_class = [each for each in abcd()]
      gate = confusion_matrix_all_class
      #pdb.set_trace() # to trace
      if len(gate) == 4 and (gate[0].indx != "4" and gate[1].indx != "3" and 
            gate[2].indx != "1" and gate[3].indx != "2"): #changed 0 to 3
        raise ValueError("confusion matrix has the wrong order of class")
      if "Micro" in self.goal:
        return micro_cal()
      else:
        return macro_cal()
    else:
      score_each_klass = [k.stats()[_goal[self.goal]] for k in
                          abcd()]
      return calculate(score_each_klass)

  def get_param(self):
    raise NotImplementedError("You should implement get_param function")


class SK_SVM(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = SVC()
    super(SK_SVM, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"kernel": ["linear", "poly", "rbf", "sigmoid"],
               "C": [0.1, 50.0],
               "coef0": [0.0, 1],
               "gamma": [0.0, 1],
               "random_state": [1, 1]}
    return tunelst

class SK_KNN(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = KNeighborsClassifier()
    super(SK_KNN, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"n_neighbors": [2,10],
               "weights": ['uniform', 'distance']}
    return tunelst

class SK_MLP(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = MLPClassifier()
    super(SK_MLP, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"hidden_layer_sizes": [100,5000],
               "activation": ['identity', 'logistic','tanh', 'relu'],
               "solver": ['lbfgs','sgd','adam'],
               "alpha": [0.0001,1],
               "learning_rate": ['constant','invscaling','adaptive'],
               "learning_rate_init": [0.001,1],
               "max_iter": [200,25000]}
    return tunelst

class SK_LDA(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = LDA()
    super(SK_LDA, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"solver": ['svd','lsqr','eigen'],
               "shrinkage": [None, 'auto'],
               "n_components": [1,4],
               "store_covariance": [True, False]}
    return tunelst

  
class SK_LSR(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = LogisticRegression()
    super(SK_LSR, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"penalty": ['l1', 'l2', 'elasticnet', 'none'],
               "C": [0.1, 50.0],
               "solver": ['newton-cg','lbfgs','liblinear','sag','saga'],
               "max_iter": [50, 1000]}
    return tunelst

class SK_DTR(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = DecisionTreeRegressor()
    super(SK_DTR, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"max_depth":[10,20],
            "min_samples_leaf":[1,10],
            "min_samples_split":[2,12]}
    return tunelst

class SK_RRF(Learners):
  def __init__(self, train_x, train_y, predict_x, predict_y, goal):
    clf = RandomForestRegressor()
    super(SK_RRF, self).__init__(clf, train_x, train_y, predict_x, predict_y,
                                 goal)

  def get_param(self):
    tunelst = {"n_estimators":[100,500],
            "max_depth":[10,20],
            "min_samples_leaf":[1,10],
            "min_samples_split":[2,12]}
    return tunelst

if __name__ == "__main__":
  pass



import shutil
from pathlib import Path

import os
import csv
import sys

import pandas as pd

from ..config import Config

DEFAULT_CONFIG = Config()

def create_folders(config=DEFAULT_CONFIG):
    folders_to_create = ["data/data_use/","results/attribute/", "results/with_CFS_DE/", "meta_data/"]

    for folder in folders_to_create:
        shutil.rmtree(folder)
        Path(folder).mkdir(parents=True, exist_ok=True)
    
    shutil.rmtree("data/data_use_{0}_months/".format(config.month))
    Path("data/data_use_{0}_months/".format(config.month)).mkdir(parents=True, exist_ok=True)

    attr_seed = os.path.join("results/attribute/", config.seed)
    Path(attr_seed).mkdir(parents=True, exist_ok=True)

    cfs_seed = os.path.join("results/with_CFS_DE/", config.seed)
    Path(cfs_seed).mkdir(parents=True, exist_ok=True)

    subfolders = ['Stats_new', 'month_{0}_models'.format(config.month)]

    for subfolder in subfolders:
        subfolder_path = os.path.join(cfs_seed, subfolder)
        if subfolder!='Stats_new':
            goals=config.goals
            for goal in goals:
                updated_goal_path=os.path.join(subfolder_path, goal)
                os.makedirs(updated_goal_path, exist_ok=True)
        else:
            os.makedirs(subfolder_path, exist_ok=True)



def copy_source_2_data_use(config=DEFAULT_CONFIG):
    source_dir = config.data_path
    dest_dir = "data/data_use/"

    for file_name in os.listdir(source_dir):
        file_path = os.path.join(source_dir, file_name)
        if file_name.endswith('.csv'):
            shutil.copy(file_path, os.path.join(dest_dir, file_name))


def rename_data_files():
    directory = "data/data_use/"
    files = os.listdir(directory)

    files.sort()

    num = 0

    record = []

    # Loop through the files and rename them
    for file in files:
        # Create the new file name with leading zeros
        new_name = '{:03d}{}'.format(num, os.path.splitext(file)[1])
        # Rename the file
        os.rename(os.path.join(directory, file), os.path.join(directory, new_name))
        # Increment the number for the next file
        num += 1
        di = {"old":os.path.splitext(file)[0], "new":new_name}
        record.append(di)
    
    pd.DataFrame.from_dict(record).to_csv("meta_data/old_new_names_mapping.csv", index=False)


def move_x_months_data(config=DEFAULT_CONFIG):
    source_folder = "data/data_use"
    destination_folder = "data/data_use_{0}_months/".format(config.month)

    for filename in os.listdir(source_folder):
        if filename.endswith(".csv"):
            filepath = os.path.join(source_folder, filename)
            df = pd.read_csv(filepath)

            num_rows = df.shape[0]

            if num_rows>config.month:
                shutil.copy(filepath, os.path.join(destination_folder, filename))

def move_back_2_data_use(config=DEFAULT_CONFIG):
    shutil.rmtree("data/data_use/")
    Path("data/data_use/").mkdir(parents=True, exist_ok=True)

    source_dir = "data/data_use_{0}_months/".format(config.month)
    dest_dir = "data/data_use/"

    cols = ["dates"] + config.goals

    # Loop through all files in the source directory
    for file_name in os.listdir(source_dir):
        file_path = os.path.join(source_dir, file_name)
        
        # Check if the file is a CSV file
        if file_name.endswith('.csv'):

            df = pd.read_csv(file_path)
            df = df[cols]
            df = df.reindex(sorted(df.columns), axis=1)

            df.to_csv(file_path, index=False)
            
            shutil.copy(file_path, os.path.join(dest_dir, file_name))


def init_datafiles(config=DEFAULT_CONFIG):
    # Create necessary folders
    create_folders(config)

    # Rename data files
    rename_data_files()

    # Only take data for x months
    move_x_months_data(config)

    # Move data back to data_use/ with some cleaning
    move_back_2_data_use(config)


import pandas as pd
import numpy as np
import math
import pickle
import sys
import utils


from ..config import Config

DEFAULT_CONFIG = Config()

def create_files(path, goal,seed, month):
    for criteria in ['mre','sa']:
        with open(path+ str(seed) + '/Stats_new/' + goal + '_' + criteria + '.txt', 'w') as f:
            results_df = pd.read_csv(path +str(seed)+'/month_' + str(month) + '_models/' + goal + '/results_tuned_final_sa.csv')
            results_df.drop(['Unnamed: 0'], axis = 1, inplace = True)
            for level in results_df.level.unique():
                sub_df = results_df[results_df['level'] == level]
                for col in ['self', 'bellwether','conv_bell','global']:
                    values = sub_df[col + '_' + criteria].values.tolist()
                    key = 'month_' + str(month) + '_level_' + str(level) + '_' + col
                    if col == 'self' + '_' + criteria:
                        key = col + '_level_' + str(level)
                    f.write("%s \n" % key)
                    for i in values:
                        f.write("%s " % i)
                    f.write("\n\n")   
                    if col == 'self':
                        seen_self = True


def generate_stats_files(config=DEFAULT_CONFIG):
    no_goals = config.no_goals
    seed = config.seed
    path = 'results/with_CFS_DE/'
    for i in range(no_goals):
        goal = utils.get_goal(i, config)
        create_files(path, goal,seed,config.month)
    


if __name__ == "__main__":
    no_goals= 7
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        path = 'results/with_CFS_DE/'
        for i in range(no_goals):
            goal = utils.get_goal(i)
            print("goal:", goal)
            create_files(path, goal,seed,12)


from src.attribute_selector import attribute_selector
from src.birch_bellwether_p_CFS import birch_bellwether_p_CFS
from src.default_bellwether import default_bellwether
from src.Find_bellwethers import find_bellwether
from src.collect_bellwethers import collect_bellwether
from src.Performance_Calculator import performance_calculator
from src.Stats_files import generate_stats_files


from __future__ import print_function, division
import random
import pdb
import time
import traceback
from utils import *

# __author__ = 'WeiFu'

class BaseSearch(object):
  def __init__(self, learner, params_distribution, train_data, tune_data,
               objective):
    self.learner = learner
    self.tune_data = tune_data
    self.params_distribution = params_distribution
    self.tune_goal = objective

  def evaluate(self):
    raise NotImplementedError("Please implement evaluate")


class DE(object):
  """
  :parameter
  ===========
  :param learner: data minier to be used to predict
  :param paras_distribution: dictionary type, key is the name, value is a
  list showing range
  :param train_data: training data sets, panda.DataFrame type
  :param tune_data: tuning data sets, panda.DataFrame type
  :param goal: tuning goal, can be "PD, PF, F, PREC, G" ect
  :param num_population: num of population in DE
  :param repeats: num of repeats,
  :param life: early termination.
  :param f: prob of mutation a+f*(b-c)
  :param cr: prob of crossover
  """

  def __init__(self, params_distribution, goal,target_class,
               num_population=10, repeats=60, f=0.75, cr=0.3, life=3):
    self.np = num_population
    self.repeats = repeats
    self.f = f
    self.cr = cr
    self.life = life
    self.target_class = str(target_class)
    self.params_distribution = params_distribution
    self.goal = goal
    self.evaluation = 0
    self.scores = {}
    self.frontier = [self.generate() for _ in range(self.np)]
    self.evaluate()
    self.bestconf, self.bestscore = self.best()

  def generate(self):
    candidate = {}
    for key, val in self.params_distribution.items():
      if isinstance(val[0], float):
        candidate[key] = round(random.uniform(val[0], val[1]), 3)
      elif isinstance(val[0], bool):
        candidate[key] = random.random() <= 0.5
      elif isinstance(val[0], str):
        candidate[key] = random.choice(val)
      elif isinstance(val[0], int):
        candidate[key] = int(random.uniform(val[0], val[1]))
      elif isinstance(val[0], list) and isinstance(val[0][0], int):
        candidate[key] = [int(random.uniform(each[0], each[1])) for each in
                          val]
      else:
        raise ValueError("type of params distribution is wrong!")
    if "random_state" in self.params_distribution.keys():
      candidate["random_state"] = 1  ## set random seed here
    return candidate

  def best(self):
    raise NotImplementedError("Please implement best")

  def evaluate(self):
    raise NotImplementedError("Please implement evaluate")

  def evaluate_once(self, **kwargs):
    raise NotImplementedError("Please implement evaluate_once")

  def get_target_score(self, score_dict):
    raise NotImplementedError("Please implement get_target_score")

  def gen3(self, n, f):
    seen = [n]

    def gen1(seen):
      while 1:
        k = random.randint(0, self.np - 1)
        if k not in seen:
          seen += [k]
          break
      return self.frontier[k]

    a = gen1(seen)
    b = gen1(seen)
    c = gen1(seen)
    return a, b, c

  def trim(self, n, x):
    if isinstance(self.params_distribution[n][0], float):
      return max(self.params_distribution[n][0],
                 min(round(x, 2), self.params_distribution[n][1]))
    elif isinstance(self.params_distribution[n][0], int):
      return max(self.params_distribution[n][0],
                 min(int(x), self.params_distribution[n][1]))
    else:
      raise ValueError("wrong type here in parameters")

  def update(self, index, old):
    newf = {}
    a, b, c = self.gen3(index, old)
    for key, val in old.items():
      if isinstance(self.params_distribution[key][0], bool):
        newf[key] = old[key] if self.cr < random.random() else not old[key]
      elif isinstance(self.params_distribution[key][0], str):
        newf[key] = random.choice(self.params_distribution[key])
      elif isinstance(self.params_distribution[key][0], list):
        temp_lst = []
        for i, each in enumerate(self.params_distribution[key]):
          temp_lst.append(old[key][i] if self.cr < random.random() else
                          max(self.params_distribution[key][i][0],
                              min(self.params_distribution[key][i][1],
                                  int(a[key][i] +
                                      self.f * (b[key][i] - c[key][i])))))
        newf[key] = temp_lst
      else:
        newf[key] = old[key] if self.cr < random.random() else self.trim(key, (
          a[key] + self.f * (b[key] - c[key])))
    return newf

  def Tune(self):
    try:
      def isBetter(new, old):
        return new < old if self.goal == 0 else new > old

      changed = False
      for k in range(self.repeats):
        #print(time.strftime("%Y%m%d_%H:%M:%S"), "###","Now life is: " ,self.life)
        if self.life <= 0:
          break
        nextgeneration = []
        for index, f in enumerate(self.frontier):
          new = self.update(index, f)
          newscore = self.get_target_score(self.evaluate_once(**new))
          self.evaluation += 1
          if isBetter(newscore[self.target_class],
                      self.scores[index][self.target_class]): 
            nextgeneration.append(new)
            self.scores[index] = newscore
          else:
            nextgeneration.append(f)
        self.frontier = nextgeneration[:]
        newbestconf, newbestscore = self.best()
        if isBetter(list(newbestscore.values()), list(self.bestscore.values())): 
          #print("newbestscore %s:" % str(newbestscore))
          #print("bestconf %s :" % str(newbestconf))
          self.bestscore = newbestscore
          self.bestconf = newbestconf
          changed = True
        if not changed:
          self.life -= 1
        changed = False
      #print("TUNING DONE !")
      return (self.bestconf, self.evaluation)
    except Exception as e:
        # Handle the exception here
        print("An error occurred in DE tune:", e)
        traceback.print_exc() 


class DE_Tune_ML(DE):
  def __init__(self, learner, params_distribution, goal, target_class,
               num_population=10, repeats=60, f=0.75, cr=0.3, life=3):
    self.learner = learner
    
    super(DE_Tune_ML, self).__init__(params_distribution, goal, target_class,
                                     num_population, repeats, f, cr, life)

  def evaluate(self):
    for n, kwargs in enumerate(self.frontier):
      score_dict = self.learner.learn({}, **kwargs)
      self.scores[n] = self.get_target_score(score_dict)

  def evaluate_once(self, **new):
    return self.learner.learn({}, **new)

  def get_target_score(self, score_dict):
    temp = {}
    for key, val in score_dict.items():
      if key == self.target_class:
        temp[key] = val  # value, not list
    return temp

  def best(self):
    # sortlst = []
    if self.goal == 0:  # the less, the better.
      sortlst = sorted(self.scores.items(),
                       key=lambda x: x[1][self.target_class], reverse=True)
    else:
      sortlst = sorted(self.scores.items(),
                       key=lambda x: x[1][self.target_class])
    bestconf = self.frontier[sortlst[-1][0]]
    bestscore = sortlst[-1][-1]
    return bestconf, bestscore





import pandas as pd
import numpy as np
import math
import pickle

from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import KFold

import platform
from os import listdir
from os.path import isfile, join
from glob import glob
from pathlib import Path
import sys
import os
import copy
import traceback
import timeit
import random


import matplotlib.pyplot as plt

import birch
from predictor_advance_v1 import *
import utils
import CFS_regression as CFS


from multiprocessing import Pool, cpu_count
from threading import Thread
from multiprocessing import Queue

# import metrices

import sys
import traceback
import warnings
warnings.filterwarnings("ignore")

from ..config import Config

DEFAULT_CONFIG = Config()

class ThreadWithReturnValue(Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None
    def run(self):
        #print(type(self._target))
        if self._target is not None:
            self._return = self._target(*self._args,
                                                **self._kwargs)
    def join(self, *args):
        Thread.join(self, *args)
        return self._return



class Bellwether(object):

    def __init__(self,data_path,attr_df, goal, month, config=DEFAULT_CONFIG):
        self.directory = data_path
        self.attr_df = attr_df
        self.cores = 8
        self.goal = goal
        self.metrics = 0
        self.month = month
        self.config = config

            
    def prepare_data(self, repo_name):
        df_raw = pd.read_csv(self.directory + repo_name, sep=',')
        df_raw = df_raw.drop(columns=['dates'])  
        last_col = utils.get_goal(self.goal, self.config)
        cols = list(df_raw.columns.values)
        cols.sort()
        cols.remove(last_col)
        df_adjust = df_raw[cols+[last_col]]
        return df_adjust

    def apply_cfs(self, df):
        goal = utils.get_goal(self.goal, self.config)
        y = df[goal].values
        X_df = df.drop(labels = [goal],axis = 1)
        X = X_df.values
        selected_cols = CFS.cfs(X,y)
        cols = X_df.columns[[selected_cols]].tolist()
        cols.append(goal)
        return df[cols],cols


    # Cluster Driver
    def cluster_driver(self,df,print_tree = True):
        X = df.apply(pd.to_numeric)
        cluster = birch.birch(branching_factor=self.config.branching_factor)
        cluster.fit(X)
        cluster_tree,max_depth = cluster.get_cluster_tree()
        if print_tree:
            print("Printing the tree")
            cluster.show_clutser_tree()
        return cluster,cluster_tree,max_depth

    def build_BIRCH(self):
        goal_name = utils.get_goal(self.goal, self.config)
        # self.attr_df = self.attr_df.drop(goal_name, axis = 1)
        # print(goal_name,self.attr_df.columns)
        cluster,cluster_tree,max_depth = self.cluster_driver(self.attr_df)
        return cluster,cluster_tree,max_depth

    
    def bellwether(self,selected_projects,all_projects):
        final_score = {}
        final_model = {}
        final_features = {}
        count = 0
        for s_project in selected_projects:
            try:
                data = self.prepare_data(s_project)
                data, cols = self.apply_cfs(data)
                list_temp, model_touse,sa = DECART_bellwether_CFS(data, self.metrics, 
                                            self.month, all_projects, s_project, 
                                            self.directory, self.goal, cols, self.config)
                final_score[s_project] = list_temp
                final_model[s_project] = model_touse
                final_features[s_project] = cols
            except ArithmeticError as e:
                print(e)
                continue
        return [final_score, final_model, final_features]

    def run_bellwether(self,projects):
        threads = []
        results = {}
        models = {}
        features = {}
        _projects = projects
        split_projects = np.array_split(_projects, self.cores)
        # split_projects = [arr for arr in split_projects if len(arr) > 0]
        for i in range(self.cores):
            print("starting thread ",i)
            t = ThreadWithReturnValue(target = self.bellwether, args = [split_projects[i],projects])
            threads.append(t)
        for th in threads:
            th.start()
        for th in threads:
            response = th.join()
            results.update(response[0])
            models.update(response[1])
            features.update(response[2])
        return results, models, features

    # For every cluster : Train model on a project and test on other projects
    # store in .pkl file
    def run(self,selected_projects,cluster_id,data_store_path):
        print(cluster_id)
        final_score, models, features = self.run_bellwether(selected_projects)
        data_path = Path(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id))
        if not data_path.is_dir():
            os.makedirs(data_path)
        with open(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id) + '/goal_' + str(self.goal)  + '.pkl', 'wb') as handle:
            pickle.dump(final_score, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
        with open(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id) + '/goal_' + str(self.goal)  + '_models.pkl', 'wb') as handle:
            pickle.dump(models, handle, protocol=pickle.HIGHEST_PROTOCOL)

        with open(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id) + '/goal_' + str(self.goal)  + '_features.pkl', 'wb') as handle:
            pickle.dump(features, handle, protocol=pickle.HIGHEST_PROTOCOL)
        # df = pd.read_pickle(data_store_path + str(cluster_id)  + '/700_RF_default_bellwether.pkl')


def birch_bellwether_p_CFS(config=DEFAULT_CONFIG):
    month = config.month
    no_goals = config.no_goals
    seed = config.seed

    for i in range(no_goals):
        print('Running Goal:', i)
        goal = utils.get_goal(i, config)
        start = timeit.default_timer()
        path = "data/data_use/"
        meta_path = 'results/attribute/'+str(seed)+'/data_attribute_' + goal + '.csv'
        data_store_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/'
        attr_df = pd.read_csv(meta_path, index_col=0)

        attr_df_index = list(attr_df.index)
        sample_size = attr_df.shape[0]

        training_projects = random.sample(attr_df_index, int(sample_size*0.8)) 
        test_projects = []
        test_len=0
        for project in attr_df_index:
            if project not in training_projects:
                test_projects.append(project) #todo use seed rand

        attr_df_train = attr_df.loc[training_projects]
        attr_df_test = attr_df.loc[test_projects]

        bell = Bellwether(path,attr_df_train,i,month,config)
        cluster,cluster_tree,max_depth = bell.build_BIRCH()

        print("Maximum depth: ", max_depth)
        print(cluster_tree)

        # Collect the leaf nodes ids in cluster_ids
        cluster_ids = []
        for key in cluster_tree:
            if cluster_tree[key].depth == max_depth:
                cluster_ids.append(key)

        # For each leaf node, run bellwether
        for ids in cluster_ids:
            selected_projects = list(attr_df_train.loc[cluster_tree[ids].data_points].index)
            print(selected_projects)
            bell.run(selected_projects,ids,data_store_path)

        data_path = Path(data_store_path + goal)
        if not data_path.is_dir():
            os.makedirs(data_path)

        attr_df_train.to_pickle(data_store_path + '/' + goal + '/train_data.pkl')
        attr_df_test.to_pickle(data_store_path + '/' + goal + '/test_data.pkl')

        stop = timeit.default_timer() 
        print("Model training time: ", stop - start)


if __name__ == "__main__":
    month = 12 #todo 12 and 24 months into the future
    no_goals= 7
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        for i in range(no_goals):
            print('Running Goal:', i)
            goal = utils.get_goal(i, DEFAULT_CONFIG)
            start = timeit.default_timer()
            path = 'data/data_use/'
            meta_path = 'results/attribute/'+str(seed)+'/data_attribute_' + goal + '.csv'
            data_store_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/'
            attr_df = pd.read_csv(meta_path, index_col=0)

            attr_df_index = list(attr_df.index)
            training_projects = random.sample(attr_df_index, 34) #124todo change to files number previously 1200
            test_projects = []
            test_len=0
            for project in attr_df_index:
                if project not in training_projects:
                    test_projects.append(project) #todo use seed rand

            attr_df_train = attr_df.loc[training_projects]
            attr_df_test = attr_df.loc[test_projects]

            bell = Bellwether(path,attr_df_train,i,month, DEFAULT_CONFIG)
            cluster,cluster_tree,max_depth = bell.build_BIRCH()

            print("Maximum depth: ", max_depth)
            print(cluster_tree)

            # Collect the leaf nodes ids in cluster_ids
            cluster_ids = []
            for key in cluster_tree:
                if cluster_tree[key].depth == max_depth:
                    cluster_ids.append(key)

            # For each leaf node, run bellwether
            for ids in cluster_ids:
                selected_projects = list(attr_df_train.loc[cluster_tree[ids].data_points].index)
                print(selected_projects)
                bell.run(selected_projects,ids,data_store_path)

            data_path = Path(data_store_path + goal)
            if not data_path.is_dir():
                os.makedirs(data_path)

            attr_df_train.to_pickle(data_store_path + '/' + goal + '/train_data.pkl')
            attr_df_test.to_pickle(data_store_path + '/' + goal + '/test_data.pkl')

            stop = timeit.default_timer() 
            print("Model training time: ", stop - start)



import pandas as pd
import numpy as np
import math
import scipy.spatial as ss
from scipy.special import digamma
from math import log
import numpy.random as nr
import random

import warnings
warnings.filterwarnings("ignore")

# continuous estimators

# def entropy(x, k=3, base=2):
#     """
#     The classic K-L k-nearest neighbor continuous entropy estimator x should be a list of vectors,
#     e.g. x = [[1.3],[3.7],[5.1],[2.4]] if x is a one-dimensional scalar and we have four samples
#     """

#     assert k <= len(x)-1, "Set k smaller than num. samples - 1"
#     print(x)
#     d = len(x[0])
#     N = len(x)
#     intens = 1e-10  # small noise to break degeneracy, see doc.
#     x = [list(p + intens * nr.rand(len(x[0]))) for p in x]
#     tree = ss.cKDTree(x)
#     nn = [tree.query(point, k+1, p=float('inf'))[0][k] for point in x]
#     const = digamma(N)-digamma(k) + d*log(2)
#     return (const + d*np.mean(map(log, nn)))/log(base)

def add_noise(x, intens=1e-10):
    # small noise to break degeneracy, see doc.
    return x + intens * np.random.random_sample(x.shape)


def query_neighbors(tree, x, k):
    return tree.query(x, k=k + 1)[0][:, k]


def count_neighbors(tree, x, r):
    return tree.query_radius(x, r, count_only=True)




def entropy(x, k=3, base=2):
    """ The classic K-L k-nearest neighbor continuous entropy estimator
        x should be a list of vectors, e.g. x = [[1.3], [3.7], [5.1], [2.4]]
        if x is a one-dimensional scalar and we have four samples
    """
    try:
        len(x[0])
    except:
        x = [[item] for item in x]
    # if(k>len(x)-1):
    #     print("CFS_REG: line 53")
    #     print(len(x))
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    x = np.asarray(x)
    n_elements, n_features = x.shape
    x = add_noise(x)
    tree = ss.cKDTree(x)
    nn = query_neighbors(tree, x, k)
    const = digamma(n_elements) - digamma(k) + n_features * log(2)
    return (const + n_features * np.log(nn).mean()) / log(base)


def mi(x, y, k=3, base=2):
    """
    Mutual information of x and y; x, y should be a list of vectors, e.g. x = [[1.3],[3.7],[5.1],[2.4]]
    if x is a one-dimensional scalar and we have four samples
    """

    assert len(x) == len(y), "Lists should have same length"
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    intens = 1e-10  # small noise to break degeneracy, see doc.
    x = [list(p + intens * nr.rand(len(x[0]))) for p in x]
    y = [list(p + intens * nr.rand(len(y[0]))) for p in y]
    points = zip2(x, y)
    # Find nearest neighbors in joint space, p=inf means max-norm
    tree = ss.cKDTree(points)
    dvec = [tree.query(point, k+1, p=float('inf'))[0][k] for point in points]
    a, b, c, d = avgdigamma(x, dvec), avgdigamma(y, dvec), digamma(k), digamma(len(x))
    return (-a-b+c+d)/log(base)


def cmi(x, y, z, k=3, base=2):
    """
    Mutual information of x and y, conditioned on z; x, y, z should be a list of vectors, e.g. x = [[1.3],[3.7],[5.1],[2.4]]
    if x is a one-dimensional scalar and we have four samples
    """

    assert len(x) == len(y), "Lists should have same length"
    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    intens = 1e-10  # small noise to break degeneracy, see doc.
    x = [list(p + intens * nr.rand(len(x[0]))) for p in x]
    y = [list(p + intens * nr.rand(len(y[0]))) for p in y]
    z = [list(p + intens * nr.rand(len(z[0]))) for p in z]
    points = zip2(x, y, z)
    # Find nearest neighbors in joint space, p=inf means max-norm
    tree = ss.cKDTree(points)
    dvec = [tree.query(point, k+1, p=float('inf'))[0][k] for point in points]
    a, b, c, d = avgdigamma(zip2(x, z), dvec), avgdigamma(zip2(y, z), dvec), avgdigamma(z, dvec), digamma(k)
    return (-a-b+c+d)/log(base)


def kldiv(x, xp, k=3, base=2):
    """
    KL Divergence between p and q for x~p(x), xp~q(x); x, xp should be a list of vectors, e.g. x = [[1.3],[3.7],[5.1],[2.4]]
    if x is a one-dimensional scalar and we have four samples
    """

    assert k <= len(x) - 1, "Set k smaller than num. samples - 1"
    assert k <= len(xp) - 1, "Set k smaller than num. samples - 1"
    assert len(x[0]) == len(xp[0]), "Two distributions must have same dim."
    d = len(x[0])
    n = len(x)
    m = len(xp)
    const = log(m) - log(n-1)
    tree = ss.cKDTree(x)
    treep = ss.cKDTree(xp)
    nn = [tree.query(point, k+1, p=float('inf'))[0][k] for point in x]
    nnp = [treep.query(point, k, p=float('inf'))[0][k-1] for point in x]
    return (const + d*np.mean(map(log, nnp))-d*np.mean(map(log, nn)))/log(base)


# Discrete estimators
def entropyd(sx, base=2):
    """
    Discrete entropy estimator given a list of samples which can be any hashable object
    """

    return entropyfromprobs(hist(sx), base=base)


def midd(x, y):
    """
    Discrete mutual information estimator given a list of samples which can be any hashable object
    """

    return -entropy(list(zip(x, y)))+entropy(x)+entropy(y)


def cmidd(x, y, z):
    """
    Discrete mutual information estimator given a list of samples which can be any hashable object
    """

    return entropy(list(zip(y, z)))+entropy(list(zip(x, z)))-entropy(list(zip(x, y, z)))-entropy(z)


def hist(sx):
    # Histogram from list of samples
    d = dict()
    for s in sx:
        d[s] = d.get(s, 0) + 1
    return map(lambda z: float(z)/len(sx), d.values())


def entropyfromprobs(probs, base=2):
    # Turn a normalized list of probabilities of discrete outcomes into entropy (base 2)
    return -sum(map(elog, probs))/log(base)


def elog(x):
    # for entropy, 0 log 0 = 0. but we get an error for putting log 0
    if x <= 0. or x >= 1.:
        return 0
    else:
        return x*log(x)


# Mixed estimators
def micd(x, y, k=3, base=2, warning=True):
    """ If x is continuous and y is discrete, compute mutual information
    """

    overallentropy = entropy(x, k, base)
    n = len(y)
    word_dict = dict()
    for sample in y:
        word_dict[sample] = word_dict.get(sample, 0) + 1./n
    yvals = list(set(word_dict.keys()))

    mi = overallentropy
    for yval in yvals:
        xgiveny = [x[i] for i in range(n) if y[i] == yval]
        if k <= len(xgiveny) - 1:
            mi -= word_dict[yval]*entropy(xgiveny, k, base)
        else:
            if warning:
                print("Warning, after conditioning, on y={0} insufficient data. Assuming maximal entropy in this case.".format(yval))
            mi -= word_dict[yval]*overallentropy
    return mi  # units already applied


# Utility functions
def vectorize(scalarlist):
    """
    Turn a list of scalars into a list of one-d vectors
    """

    return [(x,) for x in scalarlist]


def shuffle_test(measure, x, y, z=False, ns=200, ci=0.95, **kwargs):
    """
    Shuffle test
    Repeatedly shuffle the x-values and then estimate measure(x,y,[z]).
    Returns the mean and conf. interval ('ci=0.95' default) over 'ns' runs, 'measure' could me mi,cmi,
    e.g. Keyword arguments can be passed. Mutual information and CMI should have a mean near zero.
    """

    xp = x[:]   # A copy that we can shuffle
    outputs = []
    for i in range(ns):
        random.shuffle(xp)
        if z:
            outputs.append(measure(xp, y, z, **kwargs))
        else:
            outputs.append(measure(xp, y, **kwargs))
    outputs.sort()
    return np.mean(outputs), (outputs[int((1.-ci)/2*ns)], outputs[int((1.+ci)/2*ns)])


# Internal functions
def avgdigamma(points, dvec):
    # This part finds number of neighbors in some radius in the marginal space
    # returns expectation value of <psi(nx)>
    N = len(points)
    tree = ss.cKDTree(points)
    avg = 0.
    for i in range(N):
        dist = dvec[i]
        # subtlety, we don't include the boundary point,
        # but we are implicitly adding 1 to kraskov def bc center point is included
        num_points = len(tree.query_ball_point(points[i], dist-1e-15, p=float('inf')))
        avg += digamma(num_points)/N
    return avg


def zip2(*args):
    # zip2(x,y) takes the lists of vectors and makes it a list of vectors in a joint space
    # E.g. zip2([[1],[2],[3]],[[4],[5],[6]]) = [[1,4],[2,5],[3,6]]
    return [sum(sublist, []) for sublist in zip(*args)]


def information_gain(f1, f2):
    """
    This function calculates the information gain, where ig(f1,f2) = H(f1) - H(f1|f2)

    Input
    -----
    f1: {numpy array}, shape (n_samples,)
    f2: {numpy array}, shape (n_samples,)

    Output
    ------
    ig: {float}
    """
    # entropy(f1)
    conditional_entropy(f1, f2)
    ig = entropy(f1) - conditional_entropy(f1, f2)
    return ig


def conditional_entropy(f1, f2):
    """
    This function calculates the conditional entropy, where ce = H(f1) - I(f1;f2)

    Input
    -----
    f1: {numpy array}, shape (n_samples,)
    f2: {numpy array}, shape (n_samples,)

    Output
    ------
    ce: {float}
        ce is conditional entropy of f1 and f2
    """

    ce = entropy(f1) - midd(f1, f2)
    return ce


def su_calculation(f1, f2):
    """
    This function calculates the symmetrical uncertainty, where su(f1,f2) = 2*IG(f1,f2)/(H(f1)+H(f2))

    Input
    -----
    f1: {numpy array}, shape (n_samples,)
    f2: {numpy array}, shape (n_samples,)

    Output
    ------
    su: {float}
        su is the symmetrical uncertainty of f1 and f2

    """

    # calculate information gain of f1 and f2, t1 = ig(f1,f2)
    t1 = information_gain(f1, f2)
    # calculate entropy of f1, t2 = H(f1)
    t2 = entropy(f1)
    # calculate entropy of f2, t3 = H(f2)
    t3 = entropy(f2)
    # su(f1,f2) = 2*t1/(t2+t3)
    su = 2.0*t1/(t2+t3)

    return su

def merit_calculation(X, y):
    """
    This function calculates the merit of X given class labels y, where
    merits = (k * rcf)/sqrt(k+k*(k-1)*rff)
    rcf = (1/k)*sum(su(fi,y)) for all fi in X
    rff = (1/(k*(k-1)))*sum(su(fi,fj)) for all fi and fj in X

    Input
    ----------
    X: {numpy array}, shape (n_samples, n_features)
        input data
    y: {numpy array}, shape (n_samples,)
        input class labels

    Output
    ----------
    merits: {float}
        merit of a feature subset X
    """

    n_samples, n_features = X.shape
    rff = 0
    rcf = 0
    for i in range(n_features):
        fi = X[:, i]
        rcf += su_calculation(fi, y)
        for j in range(n_features):
            if j > i:
                fj = X[:, j]
                rff += su_calculation(fi, fj)
    rff *= 2
    merits = rcf / np.sqrt(n_features + rff)
    return merits


def cfs(X, y):
    """
    This function uses a correlation based heuristic to evaluate the worth of features which is called CFS

    Input
    -----
    X: {numpy array}, shape (n_samples, n_features)
        input data
    y: {numpy array}, shape (n_samples,)
        input class labels

    Output
    ------
    F: {numpy array}
        index of selected features

    Reference
    ---------
    Zhao, Zheng et al. "Advancing Feature Selection Research - ASU Feature Selection Repository" 2010.
    """

    n_samples, n_features = X.shape
    F = []
    # M stores the merit values
    M = []
    while True:
        merit = -100000000000
        idx = -1
        for i in range(n_features):
            if i not in F:
                # print('before adding', i, F)
                F.append(i)
                # calculate the merit of current selected features
                t = merit_calculation(X[:, F], y)
                if t > merit:
                    merit = t
                    idx = i
                F.pop()
        F.append(idx)
        M.append(merit)
        if len(M) > 5:
            if M[len(M)-1] <= M[len(M)-2]:
                if M[len(M)-2] <= M[len(M)-3]:
                    if M[len(M)-3] <= M[len(M)-4]:
                        if M[len(M)-4] <= M[len(M)-5]:
                            break
        if len(F) == n_features:
            break
    return np.array(F)


from sklearn.model_selection import KFold
from sklearn.utils import shuffle
from sklearn import preprocessing
import pandas as pd
import numpy as np

import os
import sys
from random import shuffle
import pickle

import json

sys.path.append('.')
from ..config import Config

DEFAULT_CONFIG = Config()

def df_split(df, month):
    trainData = df.iloc[:-month]
    testData = df.iloc[-month:]
    # if len(trainData) == 0:
    #     trainData=testData
    yield trainData, testData

# def df_split_month(df, month):
#     trainData = df.iloc[:-month]
#     testData = df.iloc[-month:]
#     return trainData, testData

def df_split_month(df, month):
    trainData = df.iloc[:-month]
    testData = df.iloc[-month:]
    if len(trainData) == 0:
        trainData=testData
    return trainData, testData


# def df_split_month(df, month):
#     if len(df) < month:
#         trainData = df
#         testData = df
#     else:
#         trainData = df.iloc[:-month]
#         testData = df.iloc[-month:]
#     return trainData, testData



def normalize(df):
    min_max_scaler = preprocessing.MinMaxScaler()
    np_scaled = min_max_scaler.fit_transform(df)
    df_normalized = pd.DataFrame(np_scaled, columns=df.columns, index=df.index)
    lst_col = df.columns[-1]
    df_normalized[lst_col] = df[lst_col]
    return df_normalized


def mre_calc(y_predict, y_actual):
    mre = []
    for predict, actual in zip(y_predict, y_actual):
        if actual == 0:
            if predict == 0:
                mre.append(0)
            elif abs(predict) <= 1:
                mre.append(1)
            else:
                mre.append(round(abs(predict - actual)+1 / (actual+1), 3))
        else:
            mre.append(round(abs(predict - actual) / (actual), 3))
    mMRE = np.median(mre)
    return mMRE


def sa_calc(Y_predict, Y_actual, X_actual):
    Absolute_Error = 0
    for predict, actual in zip(Y_predict, Y_actual):
        Absolute_Error += abs(predict - actual)
    Mean_Absolute_Error = Absolute_Error / (len(Y_predict))
    random_guess = np.mean(X_actual)
    AE_random_guess = 0
    for predict in Y_predict:
        AE_random_guess += abs(predict - random_guess)
    MAE_random_guess = AE_random_guess / (len(Y_predict))
    if MAE_random_guess == 0:
        sa_error = round((1 - (Mean_Absolute_Error+1) / (MAE_random_guess+1)), 3)
    else:
        sa_error = round((1 - Mean_Absolute_Error / MAE_random_guess), 3)
    return sa_error


def confusion_matrix_calc(TN, FP, FN, TP):
    Accuracy = (TP + TN) / (TP + FP + FN + TN)
    if TP + FP == 0:
        Precision = 0
    else:
        Precision = TP / (TP + FP)
    if TP + FN == 0:
        Recall = 0
    else:
        Recall = TP / (TP + FN)
    if Recall + Precision == 0:
        F1 = 0
    else:
        F1 = 2 * (Recall * Precision) / (Recall + Precision)
    return [Accuracy, Precision, Recall, F1]


def data_goal_arrange(repo_name, directory, goal, config):
    print(directory,repo_name)
    df_raw = pd.read_csv(directory + repo_name, sep=',')
    df_raw = df_raw.drop(columns=['dates'])
    last_col = get_goal(goal, config)
    cols = list(df_raw.columns.values)
    cols.pop(cols.index(last_col))
    df_adjust = df_raw[cols+[last_col]]

    return df_adjust


def get_goal(goal, config=DEFAULT_CONFIG):
    if goal < config.no_goals:
        return DEFAULT_CONFIG.goals[goal]
    # if goal == 0:
    #     last_col = 'monthly_contributors'
    # elif goal == 1:
    #     last_col = 'monthly_commits'
    # elif goal == 2:
    #     last_col = 'monthly_open_PRs'
    # elif goal == 3:
    #     last_col = 'monthly_closed_PRs'
    # elif goal == 4:
    #     last_col = 'monthly_open_issues'
    # elif goal == 5:
    #     last_col = 'monthly_closed_issues'
    # elif goal == 6:
    #     last_col = 'monthly_stargazer'
    # elif goal == 7:
    #     last_col = 'monthly_features'
    # elif goal == 8:
    #     last_col = 'expert_score'
    # elif goal == 9:
    #     last_col = 'license'
    # elif goal == 10:
    #     last_col = 'monthly_buggy_commits'
    # elif goal == 11:
    #     last_col = 'developer_skill'
    # # elif goal == 12:
    #     last_col = 'monthly_forks'
    # elif goal == 13:
    #     last_col = 'monthly_PR_comments'
    # elif goal == 14:
    #     last_col = 'monthly_PR_mergers'
    # elif goal == 15:
    #     last_col = 'monthly_issue_comments'
    # elif goal == 16:
    #     last_col = 'monthly_merged_PRs'
    # elif goal == 17:
    #     last_col = 'monthly_watchers'
    else:
        raise Exception("Sorry, wrong goal was passed")
    return DEFAULT_CONFIG.goals[DEFAULT_CONFIG.no_goals-1]


def df_anomaly_label(df):

    last_col_original = df.iloc[:, -1]
    # print(last_col_original)
    df_new = df.copy()
    for i in range(len(df)):
        temp = last_col_original.iloc[:i+1]
        df_new.iloc[i, -1] = np.std(temp)
    sd_column = df_new.iloc[:,-1]
    for i in range(len(sd_column)):
        if i < 12:
            sd_column[i] = 0
    # print(sd_column)
    sd_gain_column = sd_column.copy()
    for i in range(len(sd_gain_column) - 1):
        if sd_column[i] == 0:
            sd_gain_column[i + 1] = 0
        else:
            sd_gain_column[i + 1] = (sd_column[i+1] - sd_column[i]) / sd_column[i]
    # print(sd_gain_column)
    sd_label = sd_gain_column.copy()
    for i in range(len(sd_label)):
        if i < 12:
            sd_label[i] = 0
        elif sd_label[i] < -0.01 or sd_label[i] > 0.1:
            sd_label[i] = 1
        else:
            sd_label[i] = 0
    # pd.set_option("max_rows", None)
    # print(sd_label)
    df_new.iloc[:, -1] = sd_label
    train_part = int(len(sd_label) * 2 / 3)
    first_anomaly = sd_label.idxmax()
    month = train_part if train_part > first_anomaly else first_anomaly
    # print(train_part, first_anomaly, month)
    trainData = df_new.iloc[:month]
    testData = df_new.iloc[month:]
    # pd.set_option("max_columns", 4)
    # print("trainData: ", trainData)
    # print("testData: ", testData)
    yield trainData, testData


def my_confusion_matrix(list_actual, list_predict):
    tn, fp, fn, tp = 0, 0, 0, 0
    for i in range(len(list_actual)):
        if list_actual[i] == list_predict[i] == 0:
            tn += 1
        if list_actual[i] == list_predict[i] == 1:
            tp += 1
        if list_actual[i] == 0 and list_predict[i] == 1:
            fp += 1
        if list_actual[i] == 1 and list_predict[i] == 0:
            fn += 1
    return tn, fp, fn, tp

def df_norm(dframe, type="normal"):
    """ Normalize a dataframe"""
    col = dframe.columns
    bugs = dframe[dframe.columns[-1]]
    if type == "min_max":
        dframe = (dframe - dframe.min()) / (dframe.max() - dframe.min() + 1e-32)
        dframe[col[-1]] = bugs
        return dframe[col]

    if type == "normal":
        dframe = (dframe - dframe.mean()) / (dframe.std() + + 1e-32)
        dframe[col[-1]] = bugs
        return dframe[col]



if __name__ == '__main__':

    # path = r'../data/data_cleaned/'
    # repo = "abp_monthly.csv"
    #
    # dataset = data_goal_arrange(repo, path, 9)
    # print(len(dataset.columns))

    # for train, test in df_split(dataset, 1):
    #     print(train)
    #     print(test)
    #     for train, test in df_split(train, 1):
    #         print(train)
    #         print(test)

    temp = [[1.1], [2.3], [3.5]]
    print(np.around(temp))
    print(np.rint(temp).astype(np.int))



from utils import *
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor
from optimizer import *
from predictor_baseline import *
import pandas as pd

from learners import SK_DTR, SK_RRF
from DE import DE_Tune_ML
import traceback

from ..config import Config
DEFAULT_CONFIG = Config()

def tune_learner(learner, train_X, train_Y, tune_X, tune_Y, goal,
                 target_class=None):
    """
    :param learner:
    :param train_X:
    :param train_Y:
    :param tune_X:
    :param tune_Y:
    :param goal:
    :param target_class:
    :return:
    """
    try:
        if not target_class:
            target_class = goal
        clf = learner(train_X, train_Y, tune_X, tune_Y, goal)
        tuner = DE_Tune_ML(clf, clf.get_param(), goal, target_class)
        return tuner.Tune()
    except Exception as e:
        # Handle the exception here
        print("An error occurred in tune_learner:", e)
        traceback.print_exc() 

def DECART(dataset, metrics, month):

    dataset = normalize(dataset)

    trainset, testset = df_split_month(dataset, month)

    train_input = trainset.iloc[:, :-1]
    train_output = trainset.iloc[:, -1]
    test_input = testset.iloc[:, :-1]
    test_output = testset.iloc[:, -1]

    validation_trainset, valdation_testset = df_split_month(trainset, month)

    validate_train_input = validation_trainset.iloc[:, :-1]
    validate_train_output = validation_trainset.iloc[:, -1]
    validate_test_input = valdation_testset.iloc[:, :-1]
    validate_test_output = valdation_testset.iloc[:, -1]

    model = [SK_DTR][0]
    params, evaluation = tune_learner(model, validate_train_input, validate_train_output, validate_test_input,
                                            validate_test_output, metrics)
    model_touse = DecisionTreeRegressor()
    model_touse.set_params(**params)

    model_touse.fit(validate_train_input, validate_train_output)

    result_list_mre = {}
    result_list_sa = {}

    test_predict = model_touse.predict(test_input)
    test_actual = test_output.values

    mre = mre_calc(test_predict, test_actual)

    return mre, model_touse





def DECART_test(dataset, metrics, month):
    try:
        dataset = normalize(dataset)

        validate_trainset, validate_testset = df_split_month(dataset, month)
        validate_train_input = validate_trainset.iloc[:, :-1]
        validate_train_output = validate_trainset.iloc[:, -1]
        validate_test_input = validate_testset.iloc[:, :-1]
        validate_test_output = validate_testset.iloc[:, -1]

        model = [SK_DTR][0]
        params, evaluation = tune_learner(model, validate_train_input, validate_train_output, validate_test_input,
                                                validate_test_output, metrics)
        model_touse = DecisionTreeRegressor()
        model_touse.set_params(**params)

        model_touse.fit(validate_train_input, validate_train_output)

        return model_touse
    except Exception as e:
        # Handle the exception here
        print("An error occurred in DECART_test:", e)
        traceback.print_exc()


def DECART_bellwether(dataset, metrics, month, all_repo, Repo, Directory, goal, config=DEFAULT_CONFIG):

    dataset = normalize(dataset)

    validate_trainset, validate_testset = df_split_month(dataset, month)

    validate_train_input = validate_trainset.iloc[:, :-1]
    validate_train_output = validate_trainset.iloc[:, -1]
    validate_test_input = validate_testset.iloc[:, :-1]
    validate_test_output = validate_testset.iloc[:, -1]

    for test_repo in all_repo:
        test_data = data_goal_arrange(test_repo, Directory, goal, config)
        test_data = normalize(test_data) 

        _validate_testset, _ = df_split_month(test_data, month)

        test_input = _validate_testset.iloc[:, :-1]
        test_output = _validate_testset.iloc[:, -1]

        validate_test_input = pd.concat([validate_test_input,test_input], axis = 0)
        validate_test_output = pd.concat([validate_test_output,test_output], axis = 0)


    # model = [SK_DTR][0]
    # params, evaluation = tune_learner(model, validate_train_input, validate_train_output, validate_test_input,
    #                                         validate_test_output, metrics)
    model_touse = RandomForestRegressor()
    model_touse.set_params()

    model_touse.fit(validate_train_input, validate_train_output)

    result_list_mre = {}
    result_list_sa = {}

    for test_repo in all_repo:
        test_data = data_goal_arrange(test_repo, Directory, goal, config)

        test_data = normalize(test_data)

        _, _validate_testset = df_split_month(test_data, month)

        test_input = _validate_testset.iloc[:, :-1]
        test_output = _validate_testset.iloc[:, -1]

        test_predict = model_touse.predict(test_input)
        test_actual = test_output.values

        if test_repo not in result_list_mre.keys():
            result_list_mre[test_repo] = []
            result_list_sa[test_repo] = []

        mre = mre_calc(test_predict, test_actual)
        sa = sa_calc(test_predict, test_actual, validate_train_output)
        
        result_list_mre[test_repo] = mre

    if metrics == 0:
        return result_list_mre, model_touse, sa
    if metrics == 1:
        return result_list_sa, model_touse, sa

def DECART_bellwether_CFS(dataset, metrics, month, all_repo, Repo, Directory, goal, cols, config=DEFAULT_CONFIG):

    dataset = normalize(dataset)

    validate_trainset, validate_testset = df_split_month(dataset, month)

    validate_train_input = validate_trainset.iloc[:, :-1]
    validate_train_output = validate_trainset.iloc[:, -1]
    validate_test_input = validate_testset.iloc[:, :-1]
    validate_test_output = validate_testset.iloc[:, -1]

    # for test_repo in all_repo:
    #     test_data = data_goal_arrange(test_repo, Directory, goal)
    #     test_data = test_data[cols]
    #     test_data = normalize(test_data) 
    #     _validate_testset, _ = df_split_month(test_data, month)

    #     test_input = _validate_testset.iloc[:, :-1]
    #     test_output = _validate_testset.iloc[:, -1]

    #     validate_test_input = pd.concat([validate_test_input,test_input], axis = 0)
    #     validate_test_output = pd.concat([validate_test_output,test_output], axis = 0)


    model = [SK_DTR][0]
    params, evaluation = tune_learner(model, validate_train_input, validate_train_output, validate_test_input,
                                            validate_test_output, metrics)
    model_touse = DecisionTreeRegressor()
    model_touse.set_params(**params)

    model_touse.fit(validate_train_input, validate_train_output)

    result_list_mre = {}
    result_list_sa = {}

    for test_repo in all_repo:
        test_data = data_goal_arrange(test_repo, Directory, goal, config)
        test_data = test_data[cols]
        test_data = normalize(test_data)

        _, _validate_testset = df_split_month(test_data, month)

        test_input = _validate_testset.iloc[:, :-1]
        test_output = _validate_testset.iloc[:, -1]

        test_predict = model_touse.predict(test_input)
        test_actual = test_output.values

        if test_repo not in result_list_mre.keys():
            result_list_mre[test_repo] = []
            result_list_sa[test_repo] = []

        mre = mre_calc(test_predict, test_actual)
        sa = sa_calc(test_predict, test_actual, validate_train_output)
        
        result_list_mre[test_repo] = mre

    if metrics == 0:
        return result_list_mre, model_touse, sa
    if metrics == 1:
        return result_list_sa, model_touse, sa


def DERFT(dataset, metrics, month):

    dataset = normalize(dataset)

    for trainset, testset in df_split(dataset, month):
        train_input = trainset.iloc[:, :-1]
        train_output = trainset.iloc[:, -1]
        test_input = testset.iloc[:, :-1]
        test_output = testset.iloc[:, -1]

    for validate_trainset, validate_testset in df_split(trainset, 1):
        validate_train_input = validate_trainset.iloc[:, :-1]
        validate_train_output = validate_trainset.iloc[:, -1]
        validate_test_input = validate_testset.iloc[:, :-1]
        validate_test_output = validate_testset.iloc[:, -1]

    def rft_builder(a, b, c, d, e):
        model = RandomForestRegressor(
            n_estimators=a,
            max_depth=b,
            min_samples_split=c,
            min_samples_leaf=d,
            max_leaf_nodes=e
        )
        model.fit(train_input, train_output)
        test_predict = model.predict(test_input)
        test_actual = test_output.values
        if metrics == 0:
            return mre_calc(test_predict, test_actual)
        if metrics == 1:
            return sa_calc(test_predict, test_actual, train_output)

    def rft_builder_future(a, b, c, d, e):
        model = RandomForestRegressor(
            n_estimators=a,
            max_depth=b,
            min_samples_split=c,
            min_samples_leaf=d,
            max_leaf_nodes=e
        )
        model.fit(validate_train_input, validate_train_output)
        validate_test_predict = model.predict(validate_test_input)
        validate_test_actual = validate_test_output.values
        if metrics == 0:
            return mre_calc(validate_test_predict, validate_test_actual)
        if metrics == 1:
            return sa_calc(validate_test_predict, validate_test_actual, validate_train_output)

    config_optimized = de(rft_builder, metrics, bounds=[(10, 100), (1, 10), (2, 20), (1, 10), (10, 20)])[1]
    # print(config_optimized[0], config_optimized[1], config_optimized[2])
    model_touse = RandomForestRegressor(
        n_estimators = config_optimized[0],
        max_depth = config_optimized[1],
        min_samples_split = config_optimized[2],
        min_samples_leaf = config_optimized[3],
        max_leaf_nodes = config_optimized[4]
    )

    model_touse.fit(train_input, train_output)
    test_predict = np.rint(model_touse.predict(test_input))
    test_actual = test_output.values

    result_list_mre = []
    result_list_sa = []
    # print("DECART", "predict", test_predict, "actual", test_actual)
    result_list_mre.append(mre_calc(test_predict, test_actual))
    result_list_sa.append(sa_calc(test_predict, test_actual, train_output))

    if metrics == 0:
        return result_list_mre
    if metrics == 1:
        return result_list_sa


def DEKNN(dataset, metrics, month):

    dataset = normalize(dataset)

    for trainset, testset in df_split(dataset, month):
        train_input = trainset.iloc[:, :-1]
        train_output = trainset.iloc[:, -1]
        test_input = testset.iloc[:, :-1]
        test_output = testset.iloc[:, -1]

    for validate_trainset, validate_testset in df_split(trainset, 1):
        validate_train_input = validate_trainset.iloc[:, :-1]
        validate_train_output = validate_trainset.iloc[:, -1]
        validate_test_input = validate_testset.iloc[:, :-1]
        validate_test_output = validate_testset.iloc[:, -1]

    def knn_builder(a, b, c):
        model = neighbors.KNeighborsRegressor(
            n_neighbors=a,
            leaf_size=b,
            p=c
        )
        model.fit(train_input, train_output)
        test_predict = model.predict(test_input)
        test_actual = test_output.values
        if metrics == 0:
            return mre_calc(test_predict, test_actual)
        if metrics == 1:
            return sa_calc(test_predict, test_actual, train_output)

    def knn_builder_future(a, b, c):
        model = neighbors.KNeighborsRegressor(
            n_neighbors=a,
            leaf_size=b,
            p=c
        )
        model.fit(validate_train_input, validate_train_output)
        validate_test_predict = model.predict(validate_test_input)
        validate_test_actual = validate_test_output.values
        if metrics == 0:
            return mre_calc(validate_test_predict, validate_test_actual)
        if metrics == 1:
            return sa_calc(validate_test_predict, validate_test_actual, validate_train_output)

    config_optimized = de(knn_builder, metrics, bounds=[(1, 5), (10, 50), (1, 3)])[1]
    # print(config_optimized[0], config_optimized[1], config_optimized[2])
    model_touse = neighbors.KNeighborsRegressor(
        n_neighbors=config_optimized[0],
        leaf_size=config_optimized[1],
        p=config_optimized[2]
    )

    model_touse.fit(train_input, train_output)
    test_predict = np.rint(model_touse.predict(test_input))
    test_actual = test_output.values

    result_list_mre = []
    result_list_sa = []
    # print("DECART", "predict", test_predict, "actual", test_actual)
    result_list_mre.append(mre_calc(test_predict, test_actual))
    result_list_sa.append(sa_calc(test_predict, test_actual, train_output))

    if metrics == 0:
        return result_list_mre
    if metrics == 1:
        return result_list_sa


if __name__ == '__main__':

    path = r'../data/data_cleaned/'
    repo = "joker_monthly.csv"

    metrics = 0  # "0" for MRE, "1" for SA
    repeats = 1
    goal = 1
    month = 1

    data = data_goal_arrange(repo, path, goal, DEFAULT_CONFIG)

    list_temp = []
    for way in range(repeats):
        list_temp.append(DECART(data, metrics, month))

    flat_list = np.array(list_temp).flatten()
    list_output = sorted(flat_list.tolist())

    print(list_output)
    print("median DECART:", np.median(list_output))

    list_temp = []
    for way in range(repeats):
        list_temp.append(CART(data, month)[metrics])

    flat_list = np.array(list_temp).flatten()
    list_output = sorted(flat_list.tolist())

    print(list_output)
    print("median CART:", np.median(list_output))




DEFAULT_GOALS = ['monthly_contributors', 'monthly_commits',  'monthly_open_PRs', 'monthly_closed_PRs', 'monthly_open_issues', 'monthly_closed_issues', 'monthly_stargazer']

class Config:
    def __init__(self, goals:list=DEFAULT_GOALS, month:int=12, data_path:str="data/data_use/", branching_factor:int=7, data_meta:str=None, domain:list=None, seed=12345) -> None:
        self.goals = goals
        self.no_goals = len(goals)
        self.month = month
        self.data_path = data_path
        self.branching_factor = branching_factor
        self.data_meta = data_meta
        self.domain = domain
        self.seed = seed
        
    def __str__(self) -> str:
        res = "Configuration:\n"
        res += "Goal: " + " ,".join(self.goals) + "\n"
        res += "Number of Goals: " + str(self.no_goals) + "\n"
        res += "Data Path: " + str(self.data_path) + "\n"
        res += "Branching Factor: " + str(self.branching_factor) + "\n"
        res += "Meta File Path: " + str(self.data_meta) + "\n"
        res += "Domain: " +  " ,".join(self.domain) + "\n"
        res += "Seed: " + str(self.seed) + "\n"

        res = "="*50 + "\n" + res + "="*50 + "\n"

        return res 



import pandas as pd
import numpy as np
import math
import pickle

from sklearn.model_selection import train_test_split
from sklearn.metrics import classification_report
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn import metrics
from sklearn.naive_bayes import GaussianNB
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import KFold

from os import listdir
from os.path import isfile, join
from glob import glob
from pathlib import Path
import sys
import os
import timeit


import matplotlib.pyplot as plt

import birch
from predictor_advance_v1 import *
import utils


from multiprocessing import  cpu_count
from threading import Thread

# import metrices

import sys
import traceback
import warnings
warnings.filterwarnings("ignore")

from ..config import Config

DEFAULT_CONFIG = Config()

class ThreadWithReturnValue(Thread):
    def __init__(self, group=None, target=None, name=None,
                 args=(), kwargs={}, Verbose=None):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None
    def run(self):
        #print(type(self._target))
        if self._target is not None:
            self._return = self._target(*self._args,**self._kwargs)
    def join(self, *args):
        Thread.join(self, *args)
        return self._return



class Bellwether(object):

    def __init__(self,data_path,attr_df, goal, month, config=DEFAULT_CONFIG):
        self.directory = data_path
        self.attr_df = attr_df
        self.cores = cpu_count()
        self.goal = goal
        self.metrics = 0
        self.month = month
        self.config = config

            
    def prepare_data(self, repo_name):
        df_raw = pd.read_csv(self.directory + repo_name, sep=',')
        df_raw = df_raw.drop(columns=['dates'])  
        last_col = utils.get_goal(self.goal, self.config)
        cols = list(df_raw.columns.values)
        cols.remove(last_col)
        df_adjust = df_raw[cols+[last_col]]
        return df_adjust


    # Cluster Driver
    def cluster_driver(self,df,print_tree = True):
        X = df.apply(pd.to_numeric)
        cluster = birch.birch(branching_factor=self.config.branching_factor)
        cluster.fit(X)
        cluster_tree,max_depth = cluster.get_cluster_tree()
        if print_tree:
            cluster.show_clutser_tree()
        return cluster,cluster_tree,max_depth

    def build_BIRCH(self):
        goal_name = utils.get_goal(self.goal, self.config)
        # self.attr_df = self.attr_df.drop(goal_name, axis = 1)
        # print(goal_name,self.attr_df.columns)
        cluster,cluster_tree,_ = self.cluster_driver(self.attr_df)
        return cluster,cluster_tree

    
    def bellwether(self,selected_projects,all_projects):
        final_score = {}
        final_model = {}
        count = 0
        for s_project in selected_projects:
            try:
                data = self.prepare_data(s_project)
                print(s_project)
                list_temp, model_touse, sa = DECART_bellwether(data, self.metrics,self.month, all_projects, s_project,self.directory, self.goal, self.config)
                final_score[s_project] = list_temp
                final_model[s_project] = model_touse
            except ArithmeticError as e:
                print(e)
                continue
        return [final_score, final_model]

    def run_bellwether(self,projects):
        threads = []
        results = {}
        models = {}
        _projects = projects
        split_projects = np.array_split(_projects, self.cores)
        for i in range(self.cores):
            print("starting thread ",i)
            t = ThreadWithReturnValue(target = self.bellwether, args = [split_projects[i],projects])
            threads.append(t)
        for th in threads:
            th.start()
        for th in threads:
            response = th.join()
            results.update(response[0])
            models.update(response[1])
        return results,models

    def run(self,selected_projects,cluster_id,data_store_path):
        print(cluster_id)
        final_score, models = self.run_bellwether(selected_projects)
        data_path = Path(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id))
        if not data_path.is_dir():
            os.makedirs(data_path)
        with open(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id) + '/goal_' + str(self.goal)  + '.pkl', 'wb') as handle:
            pickle.dump(final_score, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
        with open(data_store_path + utils.get_goal(self.goal, self.config) + '/' + str(cluster_id) + '/goal_' + str(self.goal)  + '_models.pkl', 'wb') as handle:
            pickle.dump(models, handle, protocol=pickle.HIGHEST_PROTOCOL)
        # df = pd.read_pickle(data_store_path + str(cluster_id)  + '/700_RF_default_bellwether.pkl')


def default_bellwether(config=DEFAULT_CONFIG):
    month = config.month
    no_goals = config.no_goals
    seed = config.seed


    for i in range(no_goals):
        print('Running Goal:', i)
        goal = utils.get_goal(i, config)
        start = timeit.default_timer()
        path = "data/data_use/"
        meta_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/train_data.pkl'
        data_store_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/'
        attr_df = pd.read_pickle(meta_path)
        project_list = list(attr_df.index)
        project_list = project_list

        threads = []
        results = {}
        models = {}
        split_projects = np.array_split(project_list, cores)

        bell = Bellwether(path,attr_df,i,month, config)
        for i in range(cores):
            print("starting thread ",i)
            t = ThreadWithReturnValue(target = bell.bellwether, args = [split_projects[i],project_list])
            threads.append(t)
        for th in threads:
            th.start()
        for th in threads:
            response = th.join()
            results.update(response[0])
            models.update(response[1])
                    
        print("default path : ",data_store_path ," ", goal,'/default_bellwether.pkl')

        with open(data_store_path + goal + '/default_bellwether.pkl', 'wb') as handle:
            pickle.dump(results, handle, protocol=pickle.HIGHEST_PROTOCOL)
        
        with open(data_store_path + goal + '/default_bellwether_models.pkl', 'wb') as handle:
            pickle.dump(models, handle, protocol=pickle.HIGHEST_PROTOCOL)


if __name__ == "__main__":
    month = 12
    no_goals= 7
    cores = cpu_count()
    seeds={
    'seeds1-1':[1836],
    'seeds2-1':[8023],
    'seeds3-1':[6241],
    'seeds4-1':[3578],
    'seeds5-1':[5347],
    'seeds6-1':[2916],
    'seeds7-1':[1021],
    'seeds8-1':[9562],
    'seeds9-1':[8115],
    'seeds10-1':[6931],
    'seeds11-1':[2365],
    'seeds12-1':[9496],
    'seeds13-1':[6284],
    'seeds14-1':[4096],
    'seeds15-1':[5732],
    'seeds1-2':[4629],
    'seeds2-2':[9774],
    'seeds3-2':[1592],
    'seeds4-2':[8865],
    'seeds5-2':[7264],
    'seeds6-2':[6489],
    'seeds7-2':[3745],
    'seeds8-2':[5207],
    'seeds9-2':[2653],
    'seeds10-2':[4815],
    'seeds11-2':[7713],
    'seeds12-2':[3127],
    'seeds13-2':[1790],
    'seeds14-2':[9012],
    'seeds15-2':[1764],
    }
    seed_selected = sys.argv[1]
    seeds=seeds[seed_selected]
    for seed in seeds:
        for i in range(no_goals):
            print('Running Goal:', i)
            goal = utils.get_goal(i, DEFAULT_CONFIG)
            start = timeit.default_timer()
            path = 'data/data_use/'
            meta_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/' + goal + '/train_data.pkl'
            data_store_path = 'results/with_CFS_DE/'+str(seed)+'/month_' + str(month) + '_models/'
            attr_df = pd.read_pickle(meta_path)
            project_list = list(attr_df.index)
            project_list = project_list

            threads = []
            results = {}
            models = {}
            split_projects = np.array_split(project_list, cores)

            bell = Bellwether(path,attr_df,i,month, DEFAULT_CONFIG)
            for i in range(cores):
                print("starting thread ",i)
                t = ThreadWithReturnValue(target = bell.bellwether, args = [split_projects[i],project_list])
                threads.append(t)
            for th in threads:
                th.start()
            for th in threads:
                response = th.join()
                results.update(response[0])
                models.update(response[1])
            
            print(results)
            
            print("default path : ",data_store_path ," ", goal,'/default_bellwether.pkl')

            with open(data_store_path + goal + '/default_bellwether.pkl', 'wb') as handle:
                pickle.dump(results, handle, protocol=pickle.HIGHEST_PROTOCOL)
            
            with open(data_store_path + goal + '/default_bellwether_models.pkl', 'wb') as handle:
                pickle.dump(models, handle, protocol=pickle.HIGHEST_PROTOCOL)









import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import scipy.spatial.distance
from scipy.cluster.hierarchy import dendrogram
#Clustering birch
from freediscovery.cluster import birch_hierarchy_wrapper
from freediscovery.cluster import Birch,BirchSubcluster
#Distance measure
from scipy.spatial.distance import euclidean
#Sklearn
from sklearn.metrics.pairwise import pairwise_distances
from sklearn.model_selection import train_test_split
from sklearn import metrics
#Learners
from sklearn.svm import SVC
from sklearn.tree import DecisionTreeClassifier


import warnings

warnings.filterwarnings("ignore")

class birch(object):

    def __init__(self,threshold=0.5,branching_factor=7,n_clusters=None,outlier_threshold=0.5):
        self.threshold = threshold
        self.branching_factor = branching_factor
        self.n_clusters = n_clusters
        self.outlier_threshold = outlier_threshold
        self.Birch_clusterer = Birch(threshold=self.threshold, branching_factor=self.branching_factor,
                                     n_clusters=self.n_clusters,compute_sample_indices=True)
    # Fitting the model with train_X
    def fit(self,data):
        self.data = data
        #self.data.drop(self.data.columns[len(self.data.columns)-1], axis=1, inplace=True)
        self.Birch_clusterer.fit(self.data)

    #Defines and builds the Cluster Feature Tree
    def get_cluster_tree(self):
        self.htree, n_clusters = birch_hierarchy_wrapper(self.Birch_clusterer)
        clusters = {}
        max_depth = 0
        for i in range(n_clusters):
            node = bcluster()
            sub_cluster = self.htree.flatten()[i]
            node.set_cluster_id(sub_cluster['cluster_id'])
            depth = sub_cluster.current_depth
            node.set_depth(depth)
            if depth > max_depth:
                max_depth = depth
            if i not in clusters.keys():
                clusters[i] = {}
            if sub_cluster.current_depth == 0:
                node.set_parent()
            else:
                node.set_parent(clusters[sub_cluster.parent['cluster_id']])
            cluster_size = sub_cluster['cluster_size']
            node.set_size(cluster_size)
            data_points = sub_cluster['document_id_accumulated']
            data_points_names = self.data.iloc[data_points].index.values.tolist()
            node.set_data_points(data_points_names)
            centroid = self.data.iloc[sub_cluster['document_id_accumulated'], :].mean(axis=0).values
            node.set_centroid(centroid)
            d1,d1_v = self.calculate_d1(centroid,data_points)
            d2 = self.calculate_d2(centroid,data_points,d1_v)
            node.add_d1(d1)
            node.add_d2(d2)
            node.calculate_threshold(self.outlier_threshold)
            clusters[i] = node
            self.cluster_tree = clusters
        return self.cluster_tree,max_depth
    
    #Calculate the d1 distance(point farthest away from centroid)
    def calculate_d1(self,centroid,data_points):
        d1 = 0
        u = centroid
        d1_v = None
        for point in data_points:
            v = point
            distance = euclidean(u,v)
            if distance>d1:
                d1 = distance
                d1_v = v
        return d1,d1_v
    
    #Calculate the d2 distance(point farthest away from d1 and its distance from centroid)
    def calculate_d2(self,centroid,data_points,d1_v):
        d2_d1 = 0
        u = d1_v
        d2_v = None
        for point in data_points:
            v = point
            distance = euclidean(u,v)
            if distance>d2_d1:
                d2_d1 = distance
                d2_v = v
        d2 = euclidean(centroid,v)
        return d2
    
    # Display's the tree
    def show_clutser_tree(self):
        self.htree.display_tree()
        
        
    # Prediction Function with height based prediction with outlier detection
    def predict(self,test_X,depth):
        predicted = []
        for test_instance in test_X.iterrows():
            test_sample = test_instance[1].values
            min_distance = float('inf')
            selected_cluster = None
            for cluster_id in self.cluster_tree:
                if self.cluster_tree[cluster_id].depth != depth:
                    continue
                u = self.cluster_tree[cluster_id].centroid
                v = np.asarray(test_sample,dtype='float64')
                distance = euclidean(u,v)
                if distance < min_distance:
                    min_distance = distance
                    selected_cluster = cluster_id
            self.cluster_tree[selected_cluster].add_test_points(test_instance[0])
            # Outlier identifier
            #if self.cluster_tree[selected_cluster].check_outlier(min_distance):
            #    self.cluster_tree[selected_cluster].add_outlier_points(test_instance[0])
            #_predicted_label = self.cluster_tree[selected_cluster].classifier.predict([test_sample])
            #self.cluster_tree[selected_cluster].add_predicted(_predicted_label)
            predicted.append(selected_cluster)
        return predicted
    
    # Model certification creator
    def certify_model(self,cluster_tree,test_y):
        for cluster_id in cluster_tree:
            if len(cluster_tree[cluster_id].test_points) == 0:
                continue
            cluster_tree[cluster_id].set_test_labels(test_y[cluster_tree[cluster_id].test_points].values)
            precision = metrics.precision_score(cluster_tree[cluster_id].test_labels, 
                                                cluster_tree[cluster_id].predicted,average='weighted')
            recall = metrics.recall_score(cluster_tree[cluster_id].test_labels, 
                                          cluster_tree[cluster_id].predicted,average='weighted')
            f1_Score = metrics.f1_score(cluster_tree[cluster_id].test_labels, 
                                        cluster_tree[cluster_id].predicted,average='weighted')
            score = {'precision': precision,'recall': recall,'f1_Score': f1_Score}
            cluster_tree[cluster_id].set_score(score)



class bcluster(object):
    
    def __init__(self):
        self.parent = None
        self.parent_id = None
        self.depth = None
        self.size = None
        self.cluster_id = None
        self.data_points = []
        self.test_points = []
        self.test_labels = []
        self.predicted = []
        self.centroid = None
        self.classifier = None
        self.outlier_model = None
        self.cluster_obj = None
        self.outlier_points = []
        self.score = []
        self.d1 = None
        self.d2 = None
        self.threshold = None
    
    def set_parent(self,parent_node=None):
        if parent_node == None:
            self.parent = None
            self.parent_id = None
        else:
            self.parent = parent_node
            self.parent_id = parent_node.cluster_id
    
    def set_depth(self,depth):
        self.depth = depth
    
    def set_size(self,size):
        self.size = size
        
    def set_cluster_id(self,cluster_id):
        self.cluster_id = cluster_id
        
    def set_data_points(self,data_points):
        self.data_points = data_points
    
    def set_test_labels(self,test_labels):
        self.test_labels = test_labels
        
    def add_test_points(self,test_point):
        self.test_points.append(test_point)
        
    def add_predicted(self,predicted):
        self.predicted.append(predicted)
    
    def set_centroid(self,centroid):
        self.centroid = centroid
        
    def set_classifier(self,classifier):
        self.classifier = classifier
        
    def set_outlier_model(self,outlier_model):
        self.outlier_model = outlier_model
        
    def set_cluster_obj(self,cluster_obj):
        self.cluster_obj = cluster_obj
        
    def add_outlier_points(self,outlier_points):
        self.outlier_points.append(outlier_points)
        
    def set_score(self,score):
        self.score = score
        
    def add_d1(self,d1):
        self.d1 = d1
        
    def add_d2(self,d2):
        self.d2 = d2
        
    def calculate_threshold(self,outlier_threshold):
        self.threshold = max(self.d1,self.d2)*outlier_threshold
        
    def check_outlier(self,distance):
        if self.threshold < distance:
            result = True
        else:
            result = False
        return result